# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime

from giza.jeerah.query import equality, inequality

def query(j, app, conf):
    query_base = "project {0} and fixVersion {1} and status {2}"

    project = conf.site.projects
    sprint = getattr(conf.sprints, conf.runstate.sprint)

    queries = [
        ('total', 'project {0} and fixVersion {1}'.format(equality(project), equality(sprint))),
        ('completed', query_base.format(equality(project), equality(sprint), equality(['Closed', 'Resolved']))),
        ('progressing', query_base.format(equality(project), equality(sprint), equality(['In Code Review', 'In Progress']))),
        ('remaining', query_base.format(equality(project), equality(sprint), equality(['Open', 'Reopened'])))
    ]

    ops = []

    for name, query in queries:
        ops.append(name)
        t = app.add('task')
        t.job = j.query
        t.args = [query]
        t.description = "{0} Jira query".format(name)

    app.run()

    return dict(zip(ops, app.results))

def report(data, conf):
    result = {
        'breakdown': { },
        'counts': { },
        'meta': {
            'projects': conf.site.projects,
            'units': conf.reporting.units,
            'sprint': getattr(conf.sprints, conf.runstate.sprint),
            'date': str(datetime.date.today())
        }
    }

    for query, issues in data.items():
        result['breakdown'][query] = { }
        result['counts'][query] = 0

        for issue in issues:
            hours = issue.fields.customfield_10855

            if hours is None:
                hours = 0

            if issue.fields.assignee is None:
                assignee = 'Unassigned'
            else:
                assignee = issue.fields.assignee.name

            if assignee not in result['breakdown'][query]:
                result['breakdown'][query][assignee] = 0

            if conf.reporting.units == 'hours':
                result['breakdown'][query][assignee] += hours
            elif conf.reporting.units == 'days':
                result['breakdown'][query][assignee] += hours / 8
            elif conf.reporting.units == 'count':
                result['breakdown'][query][assignee] += 1

    for category in result['breakdown']:
        result['counts'][category] += sum(result['breakdown'][category].values())

    return result



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path

from giza.tools.strings import hyph_concat
from giza.tools.files import copy_if_needed, create_link, tarball

def html_tarball(builder, conf):
    copy_if_needed(os.path.join(conf.paths.projectroot,
                                conf.paths.includes, 'hash.rst'),
                   os.path.join(conf.paths.projectroot,
                                conf.paths.branch_output,
                                builder, 'release.txt'))

    basename = os.path.join(conf.paths.projectroot,
                            conf.paths.public_site_output,
                            hyph_concat(conf.project.name, ) + '-' + conf.git.branches.current)

    tarball_name = basename + '.tar.gz'

    tarball(name=tarball_name,
            path=builder,
            cdir=os.path.join(conf.paths.projectroot,
                              conf.paths.branch_output),
            newp=os.path.basename(basename))


    link_name = os.path.join(conf.paths.projectroot,
                             conf.paths.public_site_output,
                             conf.project.name + '.tar.gz')

    if os.path.exists(link_name):
        os.remove(link_name)

    create_link(input_fn=os.path.basename(tarball_name),
                 output_fn=link_name)

def slides_tarball(builder, conf):
    copy_if_needed(os.path.join(conf.paths.projectroot,
                                conf.paths.includes, 'hash.rst'),
                   os.path.join(conf.paths.projectroot,
                                conf.paths.branch_output,
                                builder, 'release.txt'))

    basename = os.path.join(conf.paths.projectroot,
                            conf.paths.public_site_output,
                            hyph_concat(conf.project.name,
                                        conf.git.branches.current,
                                        builder))

    tarball_fn = basename + '.tar.gz'

    tarball(name=tarball_fn,
            path=builder,
            cdir=os.path.join(conf.paths.projectroot,
                              conf.paths.branch_output),
            newp=os.path.basename(basename))

    link_name = os.path.join(conf.paths.projectroot,
                             conf.paths.public_site_output,
                             hyph_concat(conf.project.name, 'slides') + '.tar.gz')

    if os.path.exists(link_name):
        os.remove(link_name)

    create_link(input_fn=os.path.basename(tarball_fn),
                 output_fn=link_name)

def man_tarball(builder, conf):
    basename = os.path.join(conf.paths.projectroot,
                            conf.paths.public_site_output,
                            'manpages-' + conf.git.branches.current)

    tarball_name = basename + '.tar.gz'
    tarball(name=tarball_name,
            path=builder,
            cdir=os.path.join(conf.paths.projectroot, conf.paths.branch_output),
            newp=conf.project.name + '-manpages')


    link_name = os.path.join(conf.paths.projectroot,
                             conf.paths.public_site_output,
                             'manpages' + '.tar.gz')

    if os.path.exists(link_name):
        os.remove(link_name)

    create_link(input_fn=os.path.basename(tarball_name),
                 output_fn=link_name)



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
:mod:`task` stores the :class:`~giza.task.Task()` and
:class:`~giza.task.MapTask()` classes which represent single units of work in
the context of a :class:`~giza.app.BuildApp()` procedure.

:mod:`task` also has a number of dependency resolution helper functions.
"""

import logging
import sys
import os.path
import collections

logger = logging.getLogger('giza.task')

from giza.config.main import ConfigurationBase
from giza.tools.files import md5_file

if sys.version_info >= (3, 0):
    basestring = str

class Task(object):
    """
    Provides a common interface for defining an operational unit of work in a
    concurrent :class:`~giza.app.BuildApp()` environment.

    With :attr:`~giza.task.Task.target` and :attr:`~giza.task.Task.dependency`
    defined, if a ``target`` file exists and was modified after the
    ``dependency`` file, the :class:`~giza.task.Task()` operation becomes a
    no-op, unless forced.
    """

    def __init__(self, job=None, description=None, target=None, dependency=None):
        """
        All arguments are optional. You can define a :class:`~giza.task.Task()`
        either upon creation, or after creation by modifying attributes.

        :param callable job: A callable object that the task will execute.

        :param string description: Describes the task. Used in error messages.

        :param string target: A file name. A path to a file that the task will create.

        :param string dependency: A file name. A path to a file that the task
           depends on. When specified, the task will only run if forced or if
           the ``depdendency`` file is newer than the target file.
        """

        self.spec = {}
        self._conf = None
        self._args = None
        if job is not None:
            self.job = job
        self.args_type = None
        self.description = description
        self.target = target
        self.dependency = dependency
        logger.debug('created task object calling {0}, for {1}'.format(job, description))
        self._task_id = None

    @property
    def task_id(self):
        if self._task_id is None:
            self._task_id = hash(str(self.job)) + hash(str(self.args))

        return self._task_id

    @property
    def dependency(self):
        return self._dependency

    @dependency.setter
    def dependency(self, value):
        self._dependency = value

    @property
    def target(self):
        return self._target

    @target.setter
    def target(self, value):
        self._target = value

    @property
    def conf(self):
        return self._conf

    @conf.setter
    def conf(self, value):
        if isinstance(value, ConfigurationBase):
            self._conf = value
        else:
            raise TypeError

    @property
    def job(self):
        return self.spec['job']

    @job.setter
    def job(self, value):
        if isinstance(value, collections.Callable):
            self.spec['job'] = value
        else:
            raise TypeError

    @property
    def args(self):
        return self._args

    @args.setter
    def args(self, value):
        if isinstance(value, dict):
            self.args_type = 'kwargs'
            self._args = value
        elif isinstance(value, (list, tuple)):
            self.args_type = 'args'
            self._args = value
        elif isinstance(value, basestring):
            self.args_type = 'args'
            self._args = [value]
        else:
            logger.critical(type(value))

    @property
    def needs_rebuild(self):
        if (self.target is None or self.dependency is None or self.conf.runstate.force is True):
            return True
        else:
            return check_dependency(self.target, self.dependency)

    def run(self):
        logger.debug('({0}) calling {1}'.format(self.task_id, self.job))
        if self.args_type == 'kwargs':
            r = self.job(**self.args)
        elif self.args_type == 'args':
            r = self.job(*self.args)
        else:
            r = self.job()

        logger.debug('completed running task {0}, {1}'.format(self.task_id, self.description))
        return r

############### Dependency Checking ###############

def check_dependency(target, dependency):
    """
    Determines if a target requires rebuilding based on it's provided
    dependency.

    :param string target: A file name.

    :param dependency: A file name or list of file names.
    :type dependency: string, list

    :returns: A boolean. If either the ``target`` or ``dependency`` doesn't
       exist, or if the ``target`` was modified more recently than the
       ``dependency`` returns ``True`` otherwise returns ``False``.

    :func:`~giza.task.check_dependency()` Accepts dependencies in the form of a
    single file name, or as a list, and will return ``True`` if *any* dependent
    file is newer than the target.
    """

    if dependency is None:
        return True

    if isinstance(target, list):
        if len(target) == 1:
            target = target[0]
        else:
            return check_multi_dependency(target, dependency)

    if os.path.exists(target) is False:
        return True

    def needs_rebuild(targ_t, dep_f):
        if targ_t < os.path.getmtime(dep_f):
            return True
        else:
            return False

    target_time = os.path.getmtime(target)
    if isinstance(dependency, list):
        for dep in dependency:
            if needs_rebuild(target_time, dep) is True:
                return True
        return False
    else:
        return needs_rebuild(target_time, dependency)

class MapTask(Task):
    """
    A variant of :class:`~giza.task.Task()` that defines a task that like the
    kind of operation that would run in a :func:`map()` function, processing the
    contents of an operable with a single function.
    """

    def __init__(self, job=None, description=None, target=None, dependency=None):
        super(MapTask, self).__init__(job=job, description=description,
                                   target=target, dependency=dependency)
        self._iter = []

    @property
    def iter(self):
        return self._iter

    @iter.setter
    def iter(self, value):
        if isinstance(value, collections.Iterable):
            self._iter = value
        else:
            raise TypeError

    def run(self):
        return map(self.job, self.iter)

def check_multi_dependency(target, dependency):
    for idx, t in enumerate(target):
        if check_dependency(t, dependency) is True:
            return True

    return False

############### Hashed Dependency Checking ###############

def check_hashed_dependency(target, dependency, dep_map, conf):
    def normalize_fn(fn):
        if not fn.startswith(conf.paths.projectroot):
            if fn.startswith(conf.paths.source):
                fn = os.path.join(conf.paths.projectroot, fn)
            if fn.startswith('/'):
                fn = os.path.join(conf.paths.projectroot,
                                  conf.paths.source,
                                  fn[1:])

        return fn

    def needs_rebuild(t, d):
        if dep_map is None:
            return check_dependency(t, d)
        elif d in dep_map:
            fn_hash = md5_file(d)
        else:
            return check_dependency(t, d)

        if dep_map[d] == fn_hash:
            return False
        else:
            return True

    if target is None or dependency is None:
        return True

    if isinstance(target, list):
        target = [ normalize_fn(t) for t in target ]
        for f in target:
            if not os.path.exists(f):
                return True
    else:
        target = normalize_fn(target)
        if not os.path.exists(target):
            return True

    if isinstance(dependency, list):
        dependency = [ normalize_fn(d) for d in dependency ]
        for dep in dependency:
            if needs_rebuild(target, dep) is True:
                return True
        return False
    else:
        dependency = normalize_fn(dependency)
        return needs_rebuild(target, dependency)



import logging

import argh

from giza.core.git import GitRepo
from giza.config.helper import fetch_config

logger = logging.getLogger('giza.operations.git')

@argh.arg('--branch', '-b', default=None, dest='git_branch')
@argh.arg('--patch', '-p', nargs='*', dest='git_objects')
@argh.arg('--signoff', '-s', default=False, action='store_true', dest='git_sign_patch')
@argh.named('am')
def apply_patch(args):
    c = fetch_config(args)

    g = GitRepo(c.paths.projectroot)

    if c.runstate.git_branch is None:
        c.runstate.git_branch = g.current_branch()

    with g.branch(c.runstate.git_branch):
        g.am(patches=c.runstate.git_objects,
             repo='/'.join(['https://github.com', c.git.remote.upstream]),
             sign=c.runstate.git_sign_patch)

@argh.arg('--branch', '-b', default=None, dest='git_branch')
@argh.named('update')
def pull_rebase(args):
    c = fetch_config(args)

    g = GitRepo(c.paths.projectroot)

    if c.runstate.git_branch is None:
        c.runstate.git_branch = g.current_branch()

    with g.branch(c.runstate.git_branch):
        g.update()

@argh.arg('--branch', '-b', default=None, dest='git_branch')
@argh.arg('--commits', '-c', nargs='*', dest='git_objects')
@argh.named('cp')
def cherry_pick(args):
    c = fetch_config(args)

    g = GitRepo(c.paths.projectroot)

    if c.runstate.git_branch is None:
        c.runstate.git_branch = g.current_branch()

    with g.branch(c.runstate.git_branch):
        g.cherry_pick(c.runstate.git_objects)

@argh.arg('--branch', '-b', default=None, dest='git_branch')
def merge(args):
    c = fetch_config(args)

    g = GitRepo(c.paths.projectroot)

    from_branch = g.current_branch()
    branch_name = str(id(c.runstate.git_branch))

    g.checkout_branch(branch_name, c.runstate.git_branch)

    try:
        g.checkout(branch_name)
        g.rebase(from_branch)
        g.checkout(from_branch)
        g.merge(c.runstate.git_branch)
        logger.info('rebased and merged {0} into {1}'.format(c.runstate.git_branch, from_branch))
    except Exception as e:
        logger.warning('error attempting to merge branch: ' + c.runstate.git_branch)
        logger.error(e)
    finally:
        if g.current_branch != from_branch:
            g.checkout(from_branch)

        g.remove_branch(branch_name, force=True)



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Steps schema is: ::

   {
     "title": <str>,
     "stepnum": <int>,
     "pre": <str>,
     "post": <str>,
     "ref": <str>,
     "action": {
                  "heading": <str>,
                  "code": <str>,
                  "language": <str>,
                  "content": <str>,
                  "pre": <str>,
                  "post": <str>
                }
   }

Notes:

- stepnum is optional. If not specified, we assume that the sequence starts at
  one. If you specify your own ``stepnum`` in one step you have to specify your
  own step number in all steps in this sequence. The script enforces order, so
  that if you specify stepnum, you don't need to specify steps in the source file in order.

- "title" and "heading" fields can optionally hold a document that contains
  both a "text" field *and* a "character" field if you need to adjust the level
  of the heading. For example: ::

    {
      "title":
        {
          "text": "name of step",
          "character": "-"
        }
    }

  Therefore "name of step" is, by MongoDB docs convention an "h2". By default,
  heading within actions are h4s and titles of steps are h3s.

- pre/post are optional. and allow you to add prefix or postfix text to a step
  or code example/action.

- Action should be either a doc or a list of docs. Their fields are optional,
  with the following notable points:

  - "language" refers to the syntax highlighting of "code," and is unused
    otherwise.

  - "content" is a paragraph, for steps that don't have code examples.

- the spec/agg format would be at least: ::

    {
       "source":
         {
           "file": <str>
           "ref":
         }
    }

  callers may specify ``description`` and ``title`` to override the source
  location. (ref needs to be modified in calling location.)

There are several situations where we raise errors because a step document is
invalid:

1. A step has both a "source" (i.e. is an included step), and an "action" field.

2. A code block contains both "content" and "code" fields.

"""

import logging
import os.path

logger = logging.getLogger('giza.content.steps')

import giza.content.helper

from giza.tools.files import expand_tree, verbose_remove
from giza.tools.serialization import ingest_yaml_list

from rstcloth.rstcloth import RstCloth

#################### rendering ####################

class InvalidStep(Exception):
    pass

class Steps(object):
    """
    take the input file, ingest content to support spec/aggregated lists, and
    store in an internal representation.
    """
    def __init__(self, fn, cache=None):
        if cache is None:
            cache = dict()

        self.source_fn = fn
        self.agg_sources = cache
        self.source_list = ingest_yaml_list(self.source_fn)
        self.source_dir = os.path.dirname(self.source_fn)
        self.source = dict()

        sort_needed = False

        for idx, step in enumerate(self.source_list):
            if 'stepnum' not in step:
                step['stepnum'] = idx+1
            else:
                sort_needed = True

            if 'source' in step or 'inherit' in step:
                if 'source' in step:
                    source_file = step['source']['file']
                    source_ref = step['source']['ref']
                elif 'inherit' in step:
                    source_file = step['inherit']['file']
                    source_ref = step['inherit']['ref']

                if source_file in self.agg_sources:
                    current_step = self.agg_sources[source_file].get_step(source_ref)
                    msg = 'resolved ref "{0}" from file "{1}" using step cache'
                    logger.debug(msg.format(source_ref, source_file))
                else:
                    msg = 'could *not* resolved ref "{0}" from file "{1}" with step cache'
                    logger.debug(msg.format(source_ref, source_file))
                    if not os.path.exists(fn):
                        msg = 'file {0} does not exist'.format(fn)
                        logger.error(msg)
                        raise InvalidStep(msg)
                    elif fn in self.agg_sources or source_file in self.agg_sources:
                        msg = 'hitting recursion issue on {0}'.format(fn)
                        logger.error(msg)
                        raise InvalidStep(msg)
                    else:
                        msg = "reading and caching step {0} from {1} and caching"
                        logger.debug(msg.format(source_ref, source_file))
                        if source_file == os.path.basename(self.source_fn):
                            current_step = self.get_step(source_ref)
                        else:
                            steps = Steps(os.path.join(self.source_dir, source_file), self.agg_sources)

                            current_step = steps.get_step(source_ref)
                            self.agg_sources[source_file] = steps
                            self.agg_sources.update(steps.agg_sources)

                        logger.debug('successfully cached {0}'.format(source_file))

                if current_step is None:
                    msg = 'Missing ref for {0}:"{1}" in step file "{2}"'.format(source_file, source_ref, os.path.basename(self.source_fn))
                    logger.error(msg)
                    raise InvalidStep(msg)

                current_step.update(step)

                self._validate_step(current_step, ['ref', 'title'])
                self.source[source_ref] = current_step
                self.source_list[idx] = current_step
            else:
                self._validate_step(step, ['ref', 'title'])
                self.source[step['ref']] = step

        if sort_needed is True:
            self.source_list.sort(key=lambda k:k['stepnum'])

    def get_step(self, ref):
        if ref in self.source:
            step = self.source[ref].copy()
            if 'stepnum' in step:
                del step['stepnum']
            return step

    def _validate_step(self, step, keys):
        "cursory checks a step object to make sure it's valid, errors otherwise."

        valid = True
        for i in keys:
            if i not in step or step[i] is None:
                valid = False
                if i == 'ref':
                    logger.error('a step in {0} is missing a ref'.format(self.source_fn))
                else:
                    logger.error('invalid step in {0}, with ref {1}, missing {2}'.format(self.source_fn, step['ref'], i))

        if valid is False:
            raise InvalidStep('invalid steps in {0}'.format(self.source_fn))

class StepsOutput(object):
    """
    Base class for rendered step form. The render() method generates the rst in
    the internal RstCloth object.
    """

    def __init__(self, steps, conf):
        if not isinstance(steps, Steps):
            raise TypeError
        else:
            self.steps = steps

        self.conf = conf
        self.current_step = 1
        self.rst = RstCloth()
        self.hook()

    def hook(self):
        self.indent = 3

    def edition_check(self, step):
        return giza.content.helper.edition_check(step, self.conf)

    @staticmethod
    def annotate_optional(step):
        if 'optional' in step and step['optional'] is True:
            if isinstance(step['title'], dict):
                step['title']['text'] = 'Optional. ' + step['title']['text']
            else:
                if 'title' in step:
                    step['title'] = 'Optional. ' + step['title']
                elif 'heading' in step:
                    step['heading'] = 'Optional. ' + step['heading']

            del step['optional']
            return step
        else:
            return step

    def render(self):
        for step in self.steps.source_list:
            if self.edition_check(step) is False:
                continue

            step = self.annotate_optional(step)
            self.heading(step)
            self.pre(step)

            self.current_step = step['stepnum']

            if 'action' in step:
                if isinstance(step['action'], list):
                    for block in step['action']:
                        self.code_step(block)
                else:
                    self.code_step(step['action'])

            self.content(step)

            self.post(step)

    def content(self, doc):
        if 'content' in doc and doc['content'] is not None:
            self.rst.content(doc['content'], wrap=False, indent=self.indent)
            self.rst.newline()

    def pre(self, doc):
        if 'pre' in doc and doc['pre'] is not None:
            self.rst.content(doc['pre'], wrap=False, indent=self.indent)
            self.rst.newline()

    def post(self, doc, code_step=False):
        if 'post' in doc and doc['post'] is not None:
            self.rst.content(doc['post'], wrap=False, indent=self.indent)
            self.rst.newline()

        if code_step is False:
            self.post_step_hook()

    def post_step_hook(self):
        pass

    def _heading(self, block, override_char=None, indent=0):
        if 'heading' in block:
            if isinstance(block['heading'], dict):
                if 'character' in block['heading']:
                    pass
                else:
                    block['heading']['character'] = override_char
            else:
                block['heading'] = { 'text': block['heading'],
                                     'character': override_char }

            if block['heading']['text'] is None:
                logger.error('step in "{0}" is missing a heading'.format(os.path.basename(self.steps.source_fn)))
                return

            self.rst.heading(text=block['heading']['text'],
                             char=block['heading']['character'],
                             indent=indent)

            self.rst.newline()

    def code_step(self, block):
        if 'code' in block and 'content' in block:
            raise InvalidStep

        if 'heading' in block:
            self.block_heading(block)

        self.pre(block)

        if 'code' in block:
            if 'language' not in block:
                block['language'] = 'none'

            if not isinstance(block['code'], list):
                block['code'] = block['code'].split('\n')

            self.rst.directive(name='code-block',
                               arg=block['language'],
                               content=block['code'],
                               indent=self.indent)
            self.rst.newline()

        if 'content' in block:
            self.content(block['content'])

        self.post(block, code_step=True)

    def key_name(self):
        key_name = os.path.splitext(os.path.basename(self.steps.source_fn))[0]
        if key_name.startswith('step-') or key_name.startswith('steps-'):
            key_name = key_name.split('-', 1)[1]

        return key_name

class PrintStepsOutput(StepsOutput):
    """
    Variant of Steps class for generating latex-appropriate content that won't
    depend on CSS.
    """

    def block_heading(self, block):
        self._heading(block, override_char='`', indent=self.indent-3)

    def hook(self):
        self.indent = 0

    def heading(self, doc):
        # self.rst.ref_target('step-{0}-{1}-{2}'.format(doc['stepnum'],
        #                                               self.key_name(),
        #                                               doc['ref']))

        self.rst.newline()

        if isinstance(doc['title'], dict):
            doc['title']['text'] = 'Step {0}: {1}'.format(doc['stepnum'],
                                                          doc['title']['text'])
        else:
            doc['title'] = 'Step {0}: {1}'.format(doc['stepnum'], doc['title'])

        self._heading(block={ 'heading': doc['title'] },
                      override_char='~',
                      indent=self.indent)

class WebStepsOutput(StepsOutput):
    """
    Variant of Steps class for generating web-content that depends on CSS
    classes.
    """

    def hook(self):
        self.indent = 3

    def block_heading(self, block):
        self._heading(block, override_char='`', indent=self.indent)
        self.rst.directive(name="class",
                           arg="step-" + str(self.current_step),
                           indent=3)
        self.rst.newline()

    def heading(self, doc):
        self.rst.directive('only', 'not latex')

        self.rst.newline()

        h_content = ('<div class="sequence-block">' '<div class="bullet-block">'
                     '<div class="sequence-step">' '{0}' '</div>' '</div>')

        self.rst.directive(name='raw',
                           arg='html',
                           content=h_content.format(doc['stepnum']),
                           indent=3)

        self.rst.newline()

        self._heading(block={ 'heading': doc['title'] },
                      override_char='~',
                      indent=3)

        self.indent = 3
        self.rst.newline()

    def post_step_hook(self):
        self.rst.directive(name='raw',
                           arg='html',
                           content='</div>'.format(),
                           indent=3)
        self.rst.newline()

def render_step_file(input_fn, output_fn, conf):
    input_fn_base = os.path.basename(input_fn)
    logger.debug('generating step file for {0}'.format(input_fn_base))
    steps = Steps(input_fn)
    logger.debug('resolved step file input for {0}'.format(input_fn_base))

    r = RstCloth()

    web_output = WebStepsOutput(steps, conf=conf)
    web_output.render()
    r.content(web_output.rst.data, indent=0, wrap=False)
    logger.debug('generated web output for {0}'.format(input_fn_base))

    r.directive('only', 'latex')
    r.newline()
    print_output = PrintStepsOutput(steps, conf=conf)
    print_output.render()
    r.content(print_output.rst.data, indent=3, wrap=False)
    logger.debug('generated print output for {0}'.format(input_fn_base))

    if output_fn is None:
        output_fn = os.path.splitext(input_fn)[0] + '.rst'

    r.write(output_fn)
    logger.debug('wrote step include at {0}'.format(output_fn))

#################### step source iterators ####################

def _get_steps_output_fn(fn, paths):
    root_name = os.path.splitext(os.path.basename(fn).split('-', 1)[1])[0] + '.rst'

    return os.path.join(paths.projectroot, paths.branch_source, 'includes', 'steps', root_name)

def step_sources(paths):
    for fn in expand_tree(os.path.join(paths.projectroot, paths.includes), ' yaml'):
        if fn.startswith(os.path.join(paths.projectroot, paths.includes, 'step')):
            yield fn

def step_outputs(conf):
    for fn in step_sources(conf.paths):
        yield _get_steps_output_fn(fn, conf.paths)

#################### step workers ####################

def steps_tasks(conf, app):
    paths = conf.paths

    for fn in expand_tree(os.path.join(paths.projectroot, paths.includes), 'yaml'):
        if fn.startswith(os.path.join(paths.projectroot, paths.includes, 'step')):
            out_fn = _get_steps_output_fn(fn, paths)

            task = app.add('task')
            task.dependency = fn
            task.target = out_fn
            task.job = render_step_file
            task.args = [fn, out_fn, conf]
            logger.debug('added task to generate step: {0}'.format(out_fn))

def steps_clean(conf, app):
    for fn in step_outputs(conf):
        task = app.add('task')
        task.job = verbose_remove
        task.args = [fn]
        task.description = 'removing {0}'.format(fn)



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from giza.jeerah.query import equality, inequality, query_link

def query(j, app, conf):
    ops = []
    query_strings = []

    timing_query_f = '"Actual Time" is EMPTY'
    fixversion_f = "fixversion is not EMPTY"
    uncategorized_f = "fixVersion is empty or component is empty"
    project = conf.site.projects
    buckets = conf.buckets.values()

    queries = [
        ('category', "project {0} and resolution = Unresolved and ({1})".format(equality(project), uncategorized_f)),
        ('timing', 'project {0} and fixVersion {1} and resolution = Unresolved and {2} and {3}'.format(equality(project), inequality(buckets), timing_query_f, fixversion_f)),
        ('combined', 'project {0} and resolution = Unresolved and (( {1} and {2} ) or ({3}))'.format(equality(project), timing_query_f, fixversion_f, uncategorized_f)),
    ]

    for name, query in queries:
        ops.append(name)
        query_strings.append(query)

        t = app.add('task')
        t.job = j.query
        t.args = [query]
        t.description = "{0} Jira query".format(name)

    app.run()

    return zip(ops, query_strings, app.results)

def report(data, conf):
    results = {
        'totals': {},
        'queries': {}
    }

    for name, query, issues in data:
        results['totals'][name] = len(issues)
        results['queries'][name] = query_link(conf.site.url, query)
        if len(issues) < 50:
            results[name] = [ '/'.join([conf.site.url, 'browse', i.key]) for i in issues ]
        else:
            results[name] = [ 'too many']

    return results



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os.path

logger = logging.getLogger('giza.content.toc')

import yaml

import giza.content.helper
from giza.tools.files import expand_tree

from rstcloth.rstcloth import RstCloth, fill
from rstcloth.table import TableData, TableBuilder, RstTable

#################### Rendering ####################

class CustomTocTree(object):
    def __init__(self, filename, conf, sort=False):
        if "ref-toc" in filename:
            self._is_ref = True
            sort = True
        else:
            self._is_ref = False

        self.spec = self._process_spec(filename, sort)

        self.conf = conf
        self.table = None
        self.contents = None
        self.dfn = None

        self.final = False

    def build_table(self):
        self.table = TableData()
        self.table.add_header(['Name', 'Description'])

    def build_dfn(self):
        self.dfn = RstCloth()
        self.dfn.directive('class', 'toc')
        self.dfn.newline()

    def build_contents(self):
        self.contents = RstCloth()
        self.contents.directive('class', 'hidden')
        self.contents.newline()
        self.contents.directive('toctree', fields=[('titlesonly', '')], indent=3)
        self.contents.newline()

    def _process_spec(self, spec, sort=False):
        o = []

        with open(spec, 'r') as f:
            data = yaml.safe_load_all(f)

            for datum in data:
                if 'description' not in datum or datum['description'] is None:
                    datum['description'] = ''

                if sort is False:
                    pass
                elif 'name' not in datum:
                    sort = False

                o.append(datum)

        if sort is True:
            o.sort(key=lambda o: o['name'])

        return o

    def finalize(self):
        if not self.final:
            for ref in self.spec:
                if giza.content.helper.edition_check(ref, self.conf) is False:
                    continue

                if self.table is not None:
                    if 'text' in ref:
                        if ref['name'] is None:
                            self.table.add_row( [ '', ref['text'] ] )
                        else:
                            self.table.add_row( [ ref['name'], ref['text'] ])
                    if 'name' in ref:
                        self.table.add_row([ ref['name'], ref['description'] ])
                    else:
                        self.table = None

                if self.contents is not None and 'file' in ref:
                    if 'name' in ref and self._is_ref is False:
                        self.contents.content("{0} <{1}>".format(ref['name'], ref['file']), 6, wrap=False, block='toc')
                    else:
                        self.contents.content(ref['file'], 6, wrap=False, block='toc')

                if self.dfn is not None:
                    if 'name' in ref:
                        text = ref['name']
                    else:
                        text = None

                    if 'level' in ref:
                        idnt = 3 * ref['level']
                    else:
                        idnt = 3

                    if 'class' in ref:
                        self.dfn.directive(name='class', arg=ref['class'], indent=idnt)
                        idnt += 3

                    if 'text' in ref:
                        if ref['name'] is None:
                            self.dfn.content(ref['text'], idnt)
                        else:
                            self.dfn.definition(ref['name'], ref['text'], indent=idnt, bold=False, wrap=False)
                    else:
                        link = self.dfn.role('doc', ref['file'], text)
                        self.dfn.definition(link, ref['description'], indent=idnt, bold=False, wrap=False)

                    self.dfn.newline()

class TocError(Exception): pass

class AggregatedTocTree(CustomTocTree):
    def __init__(self, filename, conf):
        self.conf = conf

        self.table = None
        self.contents = None
        self.dfn = None
        self.final = False

        self.spec = []

        dfn_dir = os.path.abspath(os.path.dirname(filename))

        with open(filename, 'r') as f:
            definition = yaml.safe_load(f)

            filter_specs = []

            for dfn in definition['files']:
                if isinstance(dfn, dict):
                    if 'edition' in dfn:
                        if dfn['edition'] != self.conf.project.edition:
                            continue

                    if 'file' in dfn:
                        filter_specs.append( (dfn['file'],  dfn['level'], True) )
                    elif 'text' in dfn:
                        if 'title' in dfn:
                            filter_specs.append( ( (dfn['title'], dfn['text']), dfn['level'], False ) )
                        else:
                            filter_specs.append( ( dfn['text'], dfn['level'], False ) )
                    else:
                        raise Exception('[ERROR] [toc]: problem with {0} in {0}'.format(dfn, filename))
                else:
                    filter_specs.append( (dfn,  1, True) )

        all_objs = {}

        self._first_source = definition['sources'][0]

        for source in definition['sources']:
            with open(os.path.join(dfn_dir, source), 'r') as f:
                objs = yaml.safe_load_all(f)

                for obj in objs:
                    all_objs[obj['file']] = obj

        for fn, level, is_file in filter_specs:
            if is_file is True:
                try:
                    obj = all_objs[fn]
                    obj['level'] = level
                    self.spec.append(obj)
                except KeyError:
                    msg = 'toc: KeyError "{0}" in file: {1}'.format(fn, filename)
                    logger.error(msg)
                    raise TocError(msg)
            else:
                # translation
                if isinstance(fn, tuple):
                    self.spec.append( { 'name': fn[0],
                                        'level': level,
                                        'text': fn[1] } )
                else:
                    self.spec.append( { 'name': None,
                                        'level': level,
                                        'text': fn } )


#################### Table of Contents Generator ####################

### Internal Methods

def _get_toc_output_dir(paths):
    return os.path.join(paths.projectroot, paths.branch_source, 'includes', 'toc')

def _get_toc_base_name(fn):
    bn = os.path.basename(fn)

    if bn.startswith('ref-toc-'):
        return os.path.splitext(bn)[0][8:]
    elif bn.startswith('toc-') or bn.startswith('ref-spec-'):
        return os.path.splitext(bn)[0][4:]

def _get_toc_output_name(name, type, paths):
    dirname = _get_toc_output_dir(paths)

    if type == 'toc':
        return os.path.join(dirname, '{0}.rst'.format(name))
    else:
        return os.path.join(dirname, '{0}-{1}.rst'.format(type, name))

def _generate_toc_tree(fn, fmt, base_name, paths, conf):
    if fmt == 'spec':
        logger.debug('generating spec toc {0}'.format(fn))

        toc = AggregatedTocTree(fn, conf)
        fmt = toc._first_source[0:3]
        toc.build_dfn()
        toc.build_table()
        toc.finalize()

        if fmt == 'ref':
            outfn = _get_toc_output_name(base_name, 'table', paths)
            t = TableBuilder(RstTable(toc.table))
            t.write(outfn)
            logger.debug('wrote spec ref-toc: '  + outfn)
        elif fmt == 'toc':
            outfn = _get_toc_output_name(base_name, 'dfn-list', paths)
            toc.dfn.write(outfn)
            logger.debug('wrote spec toc: '  + outfn)
    else:
        logger.debug('generating toc {0}'.format(fn))

        toc = CustomTocTree(fn, conf)
        toc.build_contents()

        if fmt == 'toc':
            toc.build_dfn()
        elif fmt == 'ref':
            toc.build_table()

        toc.finalize()

        outfn = _get_toc_output_name(base_name, 'toc', paths)
        toc.contents.write(outfn)
        logger.debug('wrote toc: '  + outfn)

        if fmt == 'ref':
            outfn = _get_toc_output_name(base_name, 'table', paths)
            t = TableBuilder(RstTable(toc.table))
            t.write(outfn)
            logger.debug('wrote ref toc: '  + outfn)
        elif fmt == 'toc':
            outfn = _get_toc_output_name(base_name, 'dfn-list', paths)
            toc.dfn.write(outfn)
            logger.debug('wrote toc file: '  + outfn)

def toc_tasks(conf, app):
    paths = conf.paths

    for fn in expand_tree(paths.includes, 'yaml'):
        if not (fn.startswith(os.path.join(paths.includes, 'toc')) or
                fn.startswith(os.path.join(paths.includes, 'ref-toc')) or
                fn.startswith(os.path.join(paths.includes, 'ref-spec'))):
            continue
        elif len(fn) >= 24:
            task = app.add('task')
            base_name = _get_toc_base_name(fn)
            target = []

            fmt = fn[20:24]
            if fmt != 'spec':
                fmt = fn[16:19]

            task.dependency = os.path.join(paths.projectroot, fn)
            task.job = _generate_toc_tree
            task.args = [fn, fmt, base_name, paths, conf]
            task.description = 'generating {0} from {1}'.format(fmt, fn)

            if fmt != 'spec':
                target.append(_get_toc_output_name(base_name, 'toc', paths))

            is_ref_spec = fn.startswith(os.path.join(os.path.dirname(fn), 'ref-spec'))

            if not is_ref_spec and (fmt == 'toc' or fmt == 'spec'):
                target.append(_get_toc_output_name(base_name, 'dfn-list', paths))
            elif fmt == 'ref' or is_ref_spec:
                target.append(_get_toc_output_name(base_name, 'table', paths))

            task.target = target

            logger.debug('added task for generating toc from {0}'.format(fn))

def toc_clean(conf):
    rm_rf(_get_toc_output_dir(conf.paths))
    logger.info('removed all generated toc artifacts.')



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import yaml

def ingest_yaml_list(*filenames):
    o = []

    for fn in filenames:
        if isinstance(fn, list):
            filenames.extend(fn)
            continue

        data = ingest_yaml(fn)

        if isinstance(data, list):
            o.extend(data)
        else:
            o.append(data)

    return o

def ingest_yaml_doc(filename, force=False):
    data = ingest_yaml_list(filename)

    if force is True or len(data) == 1:
        return data[0]
    else:
        if len(data) == 0:
            return {}
        elif len(data) > 1:
            raise Exception('{0} has more than one document.'.format(filename))
        else:
            return data[0]

def ingest_yaml(filename):
    o = []
    with open(filename, 'r') as f:
        data = yaml.load_all(f)

        for i in data:
            o.append(i)

    if len(o) == 1:
        o = o[0]

    return o

def write_yaml(input, filename):
    with open(filename, 'w') as f:
        if isinstance(input, list):
            f.write(yaml.safe_dump_all(input, default_flow_style=False))
        elif isinstance(input, dict):
            f.write(yaml.safe_dump(input, default_flow_style=False))
        else:
            raise Exception('cannot dump $s objects to yaml.' % str(type(input)))

def dict_from_list(key, source):
    return dict( (item[key], item) for item in source )

def ingest_json(filename):
    o = []
    with open(filename, 'r') as f:
        for doc in f.readlines():
            o.append(json.loads(doc))

    if len(o) == 1:
        o = o[0]

    return o

def ingest_json_list(filename):
    o = ingest_json(filename)

    if isinstance(o, list):
        return o
    else:
        return [o]



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import logging
import time
import datetime
import tempfile
import shutil
import os
import codecs

from giza.tools.files import expand_tree

'''
This module contains utility functions used through the translate section of
giza that can obviously be used anywhere else
'''
logger = logging.getLogger('giza.translate.utils')

def get_file_list(path, input_extension):
    ''' This function wraps around expand tree to return a list of only 1 file
    if the user gives a path to a file and not a directory. Otherwise it has
    the same functionality
    :param string path: path to the file
    :param list input_extension: a list (or a single) of extensions that is acceptable
    '''
    if os.path.isfile(path):
        if input_extension is not None:
            if isinstance(input_extension, list):
                if os.path.splitext(path)[1][1:] not in input_extension:
                    return []
            else:
                if not path.endswith(input_extension):
                    return []
        return [path]
    else:
        return expand_tree(path, input_extension)


def set_logger(lg, logger_id):
    '''This method sets the formatter to the logger to have a custom field
    called the logger_id
    :param logger logger: The logger for the module
    :param string logger_id: the identifier for the instance of the module
    '''
    for handler in logger.handlers:
        lg.removeHandler(handler)

    f = logging.Formatter("%(levelname)s|%(asctime)s|%(name)s|{0}: %(message)s".format(logger_id))
    h = logging.StreamHandler(sys.stdout)
    h.setFormatter(f)
    lg.addHandler(h)
    lg.propagate = False


class Timer(object):
    '''This class is responsible for timing processes and then both logging
    them and saving them to the process's dictionary object
    '''
    def __init__(self, d, name=None, lg=logger):
        self.d = d
        self.lg = lg
        if name is None:
            self.name = 'task'
        else:
            self.name = name

    def __enter__(self):
        self.start = time.time()
        time_now = datetime.datetime.now().strftime("%Y-%m-%d %H:%M")

        self.d[self.name+"_start_time"] = time_now

        message = '[timer]: {0} started at {1}'
        message = message.format(self.name, time_now)

        self.lg.info(message)

    def __exit__(self, *args):
        total_time = time.time()-self.start
        message = '[timer]: time elapsed for {0} was: {1}'
        message = message.format(self.name, str(datetime.timedelta(seconds=total_time)))
        self.lg.info(message)

        self.d[self.name+"_time"] = total_time
        self.d[self.name+"_time_hms"] = str(datetime.timedelta(seconds=total_time))


def merge_files(output_file, input_files, annotation_list):
    '''This function merges all of the files in the file_list into the output file.
    Annotations are made in order to help differentiate which line is from
    which file. It prints out each file, interlacing their lines so you can
    compare them line by line.

    :param string output_file: The file path to output the lines to, if None goes to stdout

    :param list input_files: The list of file names to merge

    :param list annotation_list: The list of annotations to use
        (``*``,``-``,``+``,``~`, etc.)
    '''

    if len(input_files) > len(annotation_list):
        logger.error("Too many files, add more annotations and retry")
        raise TypeError("Too many files, add more annotations and retry")

    if output_file is None:
        out = sys.stdout
    else:
        out = open(output_file, 'w', 1)

    open_files = []

    for file in input_files:
        open_files.append(open(file, "r"))

    t = True
    while t:
        for index, file in enumerate(open_files):
            line = file.readline()
            if not line:
                t = False
                break
            if line[-1] == '\n':
                out.write(annotation_list[index] + line)
            else:
                out.write(annotation_list[index] + line + '\n')
        out.write("\n")

    for file in open_files:
        file.close()

    out.close()


class TempDir(object):
    ''' This class creates a temporary folder in which to put temporary files.
    It removes them automatically upon leaving the context
    '''
    def __init__(self, dir=None, super_temp=None):
        ''' This constructs the TempDir object
        :param string dir: a directory in which to put the temporary directory in
        :param string super_temp: If you have a TempDir context inside of a TempDir context, this allows you to not create two. Just pass in the directory of the previous temporary directory
        '''
        self.dir = dir
        self.super_temp = super_temp

    def __enter__(self):
        if self.super_temp is not None:
            return self.super_temp
        self.temp_dir = tempfile.mkdtemp(dir=self.dir)
        return self.temp_dir

    def __exit__(self, *args):
        if self.super_temp is None:
            shutil.rmtree(self.temp_dir, ignore_errors=True)


def flip_text_direction(in_fp, out_fp):
    ''' This function reverses every line in a file, which is helpful
    for translating text in languages from right to left where the
    model needs to compare any text in the same direction
    :param string in_fp: file path for the file to flip
    :param string out_fp: file path for the flipped file
    '''
    with codecs.open(out_fp, "w", encoding="utf-8") as out_file:
        with codecs.open(in_fp, "r", encoding="utf-8") as in_file:
            for line in in_file:
                if line[-1] == '\n':
                    out_file.write(line[-2::-1])
                else:
                    out_file.write(line[::-1])
                out_file.write('\n')



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging

logger = logging.getLogger('giza.content.primer')

from giza.tools.files import copy_if_needed, copy_always, expand_tree, verbose_remove
from giza.tools.serialization import ingest_yaml_list
from giza.tools.transformation import post_process_tasks, truncate_file, append_to_file

def get_migration_specifications(conf):
    return [ fn for fn in expand_tree(os.path.join(conf.paths.projectroot,
                                                   conf.paths.builddata))
             if  conf.project.name in os.path.basename(fn) and 'migrations' in fn ]

def directory_expansion(source_path, page, conf):
    new_page = { 'sources': expand_tree(source_path, None)}
    del page['source']

    if 'source_dir' in page:
        new_page['source_dir'] = page['source_dir']

    pages = []

    for p in convert_multi_source(new_page):
        p.update(page)
        p['target'] = os.path.join(conf.paths.projectroot,
                                   conf.paths.source,
                                   p['target'],
                                   p['source'][len(source_path)+1:])

        pages.append(p)


    return pages

def convert_multi_source(page):
    pages = [ { 'source': source }
              for source in page['sources'] ]

    for k in ('source', 'sources'):
        if k in page:
            del page[k]

    r = []
    for p in pages:
        p.update(page)
        r.append(p)

    return r


########## Path Normalization Helper Functions

def fix_migration_paths(page):
    if 'target' not in page:
        page['target'] = page['source']

    if 'override' not in page:
        page['override'] = True

    if page['target'].endswith('.txt') and page['override'] is True:
        msg = '({0}) imported files cannot end with ".txt", changing to ".rst"'
        logger.warning(msg.format(page['source']))
        page['target'] = page['target'].replace('.txt', '.rst')

    return page

def trim_leading_slash_from_pages(page):
    for field  in ['source', 'target']:
        if page[field].startswith('/'):
            page[field] = page[field][1:]

    return page

def resolve_page_path(page, conf):
    if page['target'].startswith('/'):
        fq_target = page['target']
    else:
        fq_target = os.path.join(conf.paths.projectroot, conf.paths.source, page['target'])

    if page['source'].startswith('/'):
        fq_source = page['source']
    elif 'source_dir' in page:
        fq_source = os.path.abspath(os.path.join(conf.paths.projectroot, page['source_dir'], page['source']))
    else:
        fq_source = os.path.abspath(os.path.join(conf.paths.projectroot, '..', 'source', page['source']))

    return fq_target, fq_source

########## Main Migration Operations and Task Generator

def primer_migration_tasks(conf, app):
    "Migrates all manual files to primer according to the spec. As needed."

    migration_paths = get_migration_specifications(conf)

    if len(migration_paths) == 0:
        return False
    else:
        migrations = ingest_yaml_list(*migration_paths)

        munge_jobs = []
        for page in migrations:
            if 'sources' in page:
                migrations.extend(convert_multi_source(page))
                continue

            page = fix_migration_paths(page)
            fq_target, fq_source = resolve_page_path(page, conf)

            if page['source'].endswith('/'):
                migrations.extend(directory_expansion(fq_source, page, conf))
                continue

            page = trim_leading_slash_from_pages(page)
            prev = build_migration_task(fq_target, fq_source, app)

            if 'truncate' in page:
                build_truncate_task(page['truncate'], fq_target, app)

            if 'transform' in page:
                prev.job = copy_always
                munge_jobs.append(build_transform_task(page['transform'], fq_target))

            if 'append' in page:
                prev.job = copy_always
                build_append_task(page, fq_target, migration_paths, app)

        post_process_tasks(app=app, tasks=munge_jobs)
        msg = 'added {0} migration jobs'.format(len(migrations))

        logger.info(msg)

        return True

def clean(conf, app):
    "Removes all migrated primer files according to the current spec."

    migration_paths = get_migration_specifications(conf)
    migrations = ingest_yaml_list(*migration_paths)

    targets = []
    for page in migrations:
        if 'sources' in page:
            migrations.extend(convert_multi_source(page))
            continue

        page = fix_migration_paths(page)

        targets.append(os.path.join(conf.paths.projectroot, conf.paths.source, page['target']))

    t = app.add('map')
    t.job = verbose_remove
    t.iter = targets
    t.description = 'clean primer migrations'

    logger.info('clean: removed {0} files'.format(len(targets)))

########## Task Creators

def build_migration_task(target, source, app):
    task = app.add('task')
    task.target = target
    task.job = copy_if_needed
    task.args = [ source, target, 'primer' ]

    return task

def build_transform_task(transform, target):
    return {
        'file': target,
        'type': 'primer-processing',
        'transform': transform
    }

def build_append_task(page, target, spec_files, app):
    task = app.add('task')
    task.target = page['target']
    task.job = append_to_file
    task.args = [ target, page['append'] ]
    task.dependency = spec_files

def build_truncate_task(truncate_spec, target, app):
    task = app.add('task')
    task.target = target
    task.job = truncate_file
    task.args = {
        'fn': target,
        'start_after': truncate_spec['start-after'] if 'start-after' in truncate_spec else None,
        'end_before': truncate_spec['end-before'] if 'end-before' in truncate_spec else None
    }



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os.path

logger = logging.getLogger('giza.config.paths')

from giza.tools.strings import hyph_concat
from giza.config.base import RecursiveConfigurationBase

class PathsConfig(RecursiveConfigurationBase):
    _option_registry = ['output', 'source', 'includes', 'images', 'buildsystem',
                        'tools', 'builddata']

    @property
    def locale(self):
        if self.conf.project.edition is not None and self.conf.project.edition != self.conf.project.name:
            return hyph_concat(self.state['locale'], self.conf.project.edition)
        else:
            return self.state['locale']

    @locale.setter
    def locale(self, value):
        self.state['locale'] = value

    @property
    def buildarchive(self):
        if 'buildarchive' not in self.state:
            self.buildarchive = None
        return os.path.join(self.output, 'archive')

    @buildarchive.setter
    def buildarchive(self, value):
        self.state['buildarchive'] = os.path.join(self.output, 'archive')

    @property
    def global_config(self):
        return os.path.join(self.buildsystem, 'data')

    @global_config.setter
    def global_config(self, value):
        logger.error('global_config is dynamically rendered')

    @property
    def projectroot(self):
        if 'projectroot' in self.state:
            return self.state['projectroot']
        else:
            cwd = os.getcwd()
            if 'config' in os.listdir(cwd):
                return cwd
            else:
                cwd_parts = cwd.split(os.path.sep)
                for idx, _ in enumerate(cwd_parts):
                    if idx == 0:
                        continue
                    path = os.path.sep.join(cwd_parts[0:-idx])

                    if 'config' in os.listdir(path):
                        return path

                return None

    @projectroot.setter
    def projectroot(self, value):
        if os.path.isdir(value):
            self.state['projectroot'] = value

    @property
    def public(self):
        if 'public' not in self.state:
            self.public = None

        return self.state['public']

    @public.setter
    def public(self, value):
        if self.conf.runstate.language in (None, 'en'):
            public_path = 'public'
        else:
            public_path = '-'.join(('public', self.conf.runstate.language))

        self.state['public'] = os.path.join(self.output, public_path)

    @property
    def branch_output(self):
        if 'branch_output' not in self.state:
            self.branch_output = None

        return self.state['branch_output']

    @branch_output.setter
    def branch_output(self, value):
        self.state['branch_output'] = os.path.join(self.output, self.conf.git.branches.current)

    @property
    def branch_source(self):
        if 'branch_source' not in self.state:
            self.branch_source = None

        return self.state['branch_source']

    @branch_source.setter
    def branch_source(self, value):
        p = os.path.join(self.branch_output, self.source)
        if (self.conf.project.edition is not None and
            self.conf.project.edition != self.conf.project.name):
            p += '-' + self.conf.project.edition

        self.state['branch_source'] = p

    @property
    def branch_staging(self):
        if 'branch_staging' not in self.state:
            self.branch_staging = None

        return self.state['branch_staging']

    @branch_staging.setter
    def branch_staging(self, value):
        self.state['branch_staging'] = os.path.join(self.public, self.conf.git.branches.current)

    @property
    def global_config(self):
        if 'global_config' not in self.state:
            self.global_config = None

        return self.state['global_config']

    @global_config.setter
    def global_config(self, value):
        self.state['global_config'] = os.path.join(self.buildsystem, 'data')

    @property
    def public_site_output(self):
        if 'public_site_output' in self.state:
            return self.state['public_site_output']
        else:
            p = [ self.conf.paths.public ]

            if (self.conf.project.edition != self.conf.project.name and
                self.conf.project.edition is not None):
                p.append(self.conf.project.edition)

            if self.conf.project.branched is True:
                p.append(self.conf.git.branches.current)
            elif (self.conf.git.branches.current != 'master' and
                  self.conf.project.edition is not None and
                  self.conf.git.branches.current != self.conf.git.branches.published[0]):
                p[-1] += '-' + self.conf.git.branches.current

            p = os.path.sep.join(p)

            self.state['public_site_output'] = p
            return self.state['public_site_output']

    @property
    def htaccess(self):
        if 'htaccess' in self.state:
            return self.state['htaccess']
        else:
            p = [ self.conf.paths.public ]

            if (self.conf.project.edition != self.conf.project.name and
                self.conf.project.edition is not None):
                p.append(self.conf.project.edition)

            p.append('.htaccess')

            p = os.path.sep.join(p)

            self.state['htaccess'] = p
            return self.state['htaccess']



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os.path

logger = logging.getLogger('giza.content.release')

from giza.tools.serialization import ingest_yaml
from rstcloth.rstcloth import RstCloth

def generate_output(builder, platform, version, release):
    """ This is the legacy version of the function used by the makefile and CLI infrastructure"""

    r = RstCloth()

    r.directive('code-block', 'sh', block='header')
    r.newline(block='header')

    if release == 'core':
        r.content('curl -O http://downloads.mongodb.org/{0}/mongodb-{1}-{2}.tgz'.format(platform, builder, version), 3, wrap=False, block='cmd')
    else:
        r.content('curl -O http://downloads.10gen.com/linux/mongodb-{0}-subscription-{1}-{2}.tgz'.format(builder, release, version), 3, wrap=False, block='cmd')
        r.content('tar -zxvf mongodb-{0}-subscription-{1}-{2}.tgz'.format(builder, release, version), 3, wrap=False, block='cmd')
        r.content('cp -R -n mongodb-{0}-subscription-{1}-{2}/ mongodb'.format(builder, release, version), 3, wrap=False, block='cmd')

    r.newline(block='footer')

    return r

def generate_release_output(builder, platform, architecture, release):
    """ This is the contemporary version of the function used by the generate.py script"""

    r = RstCloth()

    r.directive('code-block', 'sh', block='header')
    r.newline(block='header')

    if architecture == 'core':
        r.content('curl -O http://downloads.mongodb.org/{0}/mongodb-{1}-{2}.tgz'.format(platform, builder, release), 3, wrap=False, block='cmd')
    else:
        r.content('curl -O http://downloads.10gen.com/linux/mongodb-{0}-subscription-{1}-{2}.tgz'.format(builder, architecture, release), 3, wrap=False, block='cmd')
        r.content('tar -zxvf mongodb-{0}-subscription-{1}-{2}.tgz'.format(builder, architecture, release), 3, wrap=False, block='cmd')
        r.content('cp -R -n mongodb-{0}-subscription-{1}-{2}/ mongodb'.format(builder, architecture, release), 3, wrap=False, block='cmd')

    r.newline(block='footer')

    return r

def generate_release_untar(builder, release):
    r = RstCloth()

    r.directive('code-block', 'sh', block='header')
    r.newline(block='header')

    r.content('tar -zxvf mongodb-{0}-{1}.tgz'.format(builder, release), 3, wrap=False, block='cmd')

    return r

def generate_release_copy(builder, release):
    r = RstCloth()

    r.directive('code-block', 'sh', block='header')
    r.newline(block='header')

    r.content('mkdir -p mongodb', 3, wrap=False, block='cmd')
    r.content('cp -R -n mongodb-{0}-{1}/ mongodb'.format(builder, release), 3, wrap=False, block='cmd')

    return r

#################### Snippets for Inclusion in Installation Guides  ####################

# generate_release_output(builder, platform, version, release)

def _generate_release_ent(rel, target, release):
    r = generate_release_output( rel['type'], rel['type'].split('-')[0], rel['system'], release )
    r.write(target)
    logger.info('wrote release info file: ' + target)

def _generate_release_core(rel, target, release):
    r = generate_release_output( rel, rel.split('-')[0], 'core', release )
    r.write(target)
    logger.info('wrote release info file: ' + target)

def _generate_untar_core(rel, target, release):
    r = generate_release_untar(rel, release)
    r.write(target)
    logger.info('wrote release info file: ' + target)

def _generate_copy_core(rel, target, release):
    r = generate_release_copy(rel, release)
    r.write(target)
    logger.info('wrote release info file: ' + target)

def release_tasks(conf, app):
    if 'releases' not in conf.system.files.data:
        return

    if 'release' in conf.version:
        release_version = conf.version.release
    else:
        release_version = conf.version.published[0]

    rel_data = conf.system.files.data.releases

    deps = [ os.path.join(conf.paths.projectroot, conf.runstate.conf_path) ]
    for rel in rel_data['source-files']:
        target = os.path.join(conf.paths.projectroot,
                              conf.paths.includes,
                              'install-curl-release-{0}.rst'.format(rel))

        t = app.add('task')
        t.job = _generate_release_core
        t.args =  [ rel, target, release_version ]
        t.target = target
        t.dependency = deps
        t.description = 'generating release page {0}'.format(target)

        target = os.path.join(conf.paths.projectroot,
                              conf.paths.includes,
                              'install-untar-release-{0}.rst'.format(rel))

        t = app.add('task')
        t.job = _generate_untar_core
        t.args = [ rel, target, release_version ]
        t.target = target
        t.dependency = deps
        t.description = 'generating release page {0}'.format(target)

        target = os.path.join(conf.paths.projectroot,
                              conf.paths.includes,
                              'install-copy-release-{0}.rst'.format(rel))

        t = app.add('task')
        t.job = _generate_copy_core
        t.args = [ rel, target, release_version ]
        t.target = target
        t.dependency = deps
        t.description = 'generating release page {0}'.format(target)

    for rel in rel_data['subscription-build']:
        target = os.path.join(conf.paths.projectroot, conf.paths.includes,
                              'install-curl-release-ent-{0}.rst'.format(rel['system']))


        t = app.add('task')
        t.job = _generate_release_ent
        t.args = [ rel, target, release_version ]
        t.target = target
        t.dependency = deps
        t.description = 'generating release page {0}'.format(target)



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Provides the :class:`~giza.git.GitRepo()` class that provides a thin
Python-layer on top of common git operations.
"""

import logging
import os
import re
from contextlib import contextmanager

from giza.tools.command import command, CommandError

logger = logging.getLogger('giza.git')

class GitError(Exception): pass

class GitRepo(object):
    """
    An object that represents a git repository, and provides methods to wrap
    many common git commands and basic aggregate operations.
    """

    def __init__(self, path=None):
        """
        :param string path: Optional. Defines a the path of the git
           repository. If not specified, defaults to the current working
           directory.
        """

        if path is None:
            self.path = os.getcwd()
        else:
            self.path = path

        logger.debug("created git repository management object for {0}".format(self.path))

    def cmd(self, *args):
        args = ' '.join(args)

        try: 
            return command(command='cd {0} ; git {1}'.format(self.path, args), capture=True)
        except CommandError as e:
            raise GitError(e)

    def remotes(self):
        return self.cmd('remote').out.split('\n')

    def branch_exists(self, name):
        r = self.cmd('branch --list ' + name).out.split('\n')

        if len(r) == 0:
            return False
        elif len(r) == 1: 
            return True
        else: 
            raise GitError

    def branch_file(self, path, branch='master'):
        return self.cmd('show {branch}:{path}'.format(branch=branch, path=path)).out

    def checkout(self, ref):
        return self.cmd('checkout', ref)

    def checkout_branch(self, name, tracking=None):
        args = ['checkout', '-b', name]

        if tracking is not None:
            args.append(tracking)

        return self.cmd(*args)

    def remove_branch(self, name, force=False):
        args = ['branch']

        if force is False:
            args.append('-d')
        else:
            args.append('-D')

        args.append(name)

        return self.cmd(*args)

    def rebase(self, onto):
        return self.cmd('rebase', onto)

    def merge(self, branch):
        return self.cmd('merge', branch)

    def hard_reset(self, ref='HEAD'):
        return self.cmd('reset', '--hard', ref)

    def reset(self, ref='HEAD'):
        return self.cmd('reset', ref)

    def fetch(self, remote='origin'):
        return self.cmd('fetch', remote)

    def update(self):
        return self.cmd('pull', '--rebase')

    def pull(self, remote='origin', branch='master'):
        return self.cmd('pull', remote, branch)

    def current_branch(self):
        return self.cmd('symbolic-ref', 'HEAD').out.split('/')[2]

    def sha(self, ref='HEAD'):
        return self.cmd('rev-parse', '--verify', ref).out

    def clone(self, remote, repo_path=None, branch=None):
        args = ['clone', remote]

        if branch is not None:
            args.extend(['--branch', branch])

        if repo_path is not None:
            args.append(repo_path)

        return self.cmd(*args)

    def cherry_pick(self, *args):
        if len(args) == 1:
            args = args[0]

        for commit in args:
            self.cmd('cherry-pick', commit)
            logger.info('cherry picked ' + commit )

    def am(self, patches, repo=None, sign=False):
        cmd_base = 'curl {path} | git am --3way'

        if sign is True:
            cmd_base += ' --signoff'

        for obj in patches:
            if obj.startswith('http'):
                if not obj.endswith('.patch'):
                    obj += '.patch'

                command(cmd_base.format(path=obj))
                logger.info("applied {0}".format(obj))
            elif re.search('[a-zA-Z]+', obj):
                path = '/'.join([ repo, 'commit', obj ]) + '.patch'

                command(cmd_base.format(path=path))
                logger.info('merged commit {0} for {1} into {2}'.format(obj, repo, self.current_branch()))
            else:
                if repo is None:
                    logger.warning('not applying "{0}", because of missing repo'.format(obj))
                else:
                    path = '/'.join([ repo, 'pull', obj ]) + '.patch'
                    command(cmd_base.format(path=path))
                    logger.info("applied {0}".format(obj))

    @contextmanager
    def branch(self, name):
        starting_branch = self.current_branch()

        if name != starting_branch:
            self.checkout(name)

        yield

        if name != starting_branch:
            self.checkout(starting_branch)



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from unittest import TestCase

from giza.core.inheritance import DataContentBase, DataCache, InheritableContentError, InheritableContentBase

from giza.config.main import Configuration
from giza.config.runtime import RuntimeStateConfig
from giza.config.base import RecursiveConfigurationBase

class DummyRecord(InheritableContentBase):
    _option_registry = ['pre', 'post', 'final', 'ref', 'title', 'content', 'edition', 'operation', 'results']
class DummyContent(DataContentBase):
    content_class = DummyRecord
class DummyCache(DataCache):
    content_class = DummyContent


def get_inheritance_data_files():
    return [
        os.path.abspath(os.path.join(os.path.dirname(__file__), 'data-inheritance', fn))
        for fn in ( 'example-add-one.yaml', 'example-add-two.yaml',
                    'example-add-three.yaml' )
    ]

class TestDataCache(TestCase):
    @classmethod
    def setUp(self):
        self.c = Configuration()
        self.c.runstate = RuntimeStateConfig()
        self.c.paths = { 'includes':
                         os.path.join(os.path.abspath(os.path.dirname(__file__)), 'data-inheritance')}

        self.data = DummyCache([], self.c)

    def test_content_class_default(self):
        self.assertEqual(self.data.cache, {})
        self.assertIs(self.data.content_class, DummyContent)

    def test_cache_not_setable(self):
        self.assertEqual(self.data.cache, {})
        self.data.cache = []
        self.assertEqual(self.data.cache, {})

    def test_cache_property_reference(self):
        self.assertIs(self.data.cache, self.data._cache)

    def test_cache_cleaner(self):
        self.assertEqual(self.data.cache, {})

        self.data._clear_cache('foo')
        self.data._clear_cache('bar')

        self.assertIn('foo', self.data.cache)
        self.assertIn('bar', self.data.cache)

        self.assertEqual(self.data.cache['foo'], [])
        self.assertEqual(self.data.cache['bar'], [])

    def test_membership(self):
        self.assertEqual(self.data.cache, {})

        self.data._clear_cache('foo')
        self.data._clear_cache('bar')

        self.assertIn('foo', self.data.cache)
        self.assertIn('bar', self.data.cache)

        self.assertIn('foo', self.data)
        self.assertIn('bar', self.data)

    def test_ingest(self):
        files = get_inheritance_data_files()
        self.assertEqual(self.data.cache, {})
        self.data.ingest(files)
        self.assertNotEqual(self.data.cache, {})

        for fn in files:
            self.assertIn(fn, self.data)

    def test_ingest_ignore_duplicates(self):
        files = get_inheritance_data_files()
        self.assertEqual(self.data.cache, {})
        self.data.ingest(files)
        self.data.ingest(files)
        self.assertNotEqual(self.data.cache, {})

        self.assertEqual(len(self.data.cache), len(files))

    def test_add_file(self):
        self.assertEqual(self.data.cache, {})
        for fn in get_inheritance_data_files():
            self.data.add_file(fn)
            self.assertIn(fn, self.data.cache)
            self.assertIn(fn, self.data)
            self.assertIsInstance(self.data.cache[fn], self.data.content_class)

    def test_fetch(self):
        files = get_inheritance_data_files()
        self.assertEqual(self.data.cache, {})
        self.data.ingest(files)

        for fn in files:
            self.assertIn(fn, self.data)
            for idx in range(len(self.data.cache), 1):
                content = self.data.fetch(fn, idx)
                self.assertIsInstance(content, dict)

        self.assertEqual(len(self.data.cache), len(files))

    def test_fetch_without_adding_file(self):
        files = get_inheritance_data_files()
        self.assertEqual(self.data.cache, {})

        for fn in files:
            with self.assertRaises(InheritableContentError):
                content = self.data.fetch(fn, 1)

            self.assertNotIn(fn, self.data)

class TestDataContentBase(TestCase):
    @classmethod
    def setUp(self):
        self.c = Configuration()
        self.c.runstate = RuntimeStateConfig()
        self.c.paths = { 'includes':
                         os.path.join(os.path.abspath(os.path.dirname(__file__)), 'data-inheritance')}

        self.content_fn = get_inheritance_data_files()[0]

        self.data = DummyCache([], self.c)
        self.data.ingest([self.content_fn])

        self.content = self.data.cache[self.content_fn]

    def test_content_created(self):
        for fn, example in self.data.cache.items():
            for v in example.content.values():
                self.assertIsInstance(v, RecursiveConfigurationBase)

    def test_content_is_correct_type(self):
        self.assertIsInstance(self.content, DataContentBase)

    def test_content_state_reference(self):
        self.assertIs(self.content.content, self.content.state['content'])

    def test_reference_to_cache(self):
        self.assertIs(self.content.data, self.data)

    def test_fetching_content(self):
        for idx in self.content.content:
            self.assertIs(self.content.content[idx], self.data.fetch(self.content_fn, idx))

    def test_resolve_checker(self):
        if isinstance(self.content, dict):
            with self.assertRaises(AttributeError):
                self.content.is_resolved()
        else:
            self.assertIsInstance(self.content.is_resolved(), bool)

class TestInheritedContentResolution(TestCase):
    @classmethod
    def setUp(self):
        self.c = Configuration()
        self.c.runstate = RuntimeStateConfig()
        self.c.paths = { 'includes':
                         os.path.join(os.path.abspath(os.path.dirname(__file__)), 'data-inheritance')}

        self.data = DummyCache(get_inheritance_data_files(), self.c)

    def test_gross_correctness_of_ingestion(self):
        self.assertEqual(len(self.data.cache), 3)

    # def test_everything_resolved(self):
    #     for fn, data in self.data.cache.items():
    #         self.assertIsInstance(data, self.data.content_class)

    #         self.assertNotEqual(len(data.content), 0)

    #         for doc in data.content.values():
    #             if 'source' in doc:
    #                 self.assertEqual(len(doc.state.keys()), 2)

    #         data.resolve()

    #         for doc in data.content.values():
    #             if 'source' in doc:
    #                 print len(doc)
    #                 self.assertTrue(doc.source.resolved)



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os.path
import logging

logger = logging.getLogger('giza.content.images')

from docutils.core import publish_parts
from rstcloth.rstcloth import RstCloth

from giza.tools.command import command
from giza.tools.files import verbose_remove
from giza.tools.serialization import ingest_yaml_list
from giza.tools.strings import dot_concat, hyph_concat

## Internal Supporting Methods

def generate_image_pages(dir, name, alt, output, conf):
    r = RstCloth()

    image = '/'.join([dir, name])
    b = name

    for img_output in output:
        img_output['width'] = str(img_output['width']) + 'px'

        r.newline()

        if 'tag' in img_output:
            tag = '-' + img_output['tag'] + '.png'
        else:
            tag = '.png'

        options = [('alt', alt), ('align', 'center'), ('figwidth', img_output['width'])]

        if 'scale' in img_output:
            options.append(('scale', img_output['scale']))

        if img_output['type'] == 'print':
            r.directive('only', 'latex', wrap=False, block=b)
            r.newline()
            r.directive(name='figure',
                        arg='/images/{0}{1}'.format(name, tag),
                        fields=options,
                        indent=3,
                        content=alt,
                        block=b)
        else:
            alt_html = publish_parts(alt, writer_name='html')['body'].strip()
            img_tags = ['<div class="figure align-center" style="max-width:{5};">',
                        '<img src="{0}/{1}/_images/{2}{3}" alt="{4}">', '</img>',
                        '{6}</div>' ]
            img_str = ''.join(img_tags)


            r.directive('only', 'website and not html', wrap=False, block=b)
            r.newline()
            r.directive(name='raw', arg='html',
                        content=img_str.format(conf.project.url,
                                               conf.git.branches.current, name, tag, alt,
                                               img_output['width'], alt_html),
                        indent=3,
                        block=b)

            r.newline(count=2)

            if img_output['width'] > 600:
                options[2] = ('figwidth', 600)

            r.directive('only', 'website and html', wrap=False, block=b)
            r.newline()
            r.directive(name='figure',
                        arg='/images/{0}{1}'.format(name, tag),
                        fields=options,
                        indent=3,
                        content=alt,
                        block=b)


        r.newline(block=b)

    image_rst_file_path = os.path.join(conf.paths.projectroot, image + '.rst')
    r.write(image_rst_file_path)
    logger.info('generated include file {0}.rst'.format(image))

def _get_inkscape_cmd():
    if sys.platform in ['linux', 'linux2']:
        return '/usr/bin/inkscape'
    elif sys.platform == 'darwin':
        inkscape = '/Applications/Inkscape.app/Contents/Resources/bin/inkscape'
        if os.path.exists(inkscape):
            return inkscape

    return 'inkscape'

def _generate_images(cmd, dpi, width, target, source):
    command(cmd.format(cmd=_get_inkscape_cmd(),
                       dpi=dpi,
                       width=width,
                       target=target,
                       source=source))
    logger.info('generated image file {0}'.format(target))

def get_images_metadata_file(conf):
    base = None
    for fn in conf.system.files.paths:
        if isinstance(fn, dict):
            if 'images' in fn:
                base = fn['images']
                break
        else:
            if fn.startswith('images'):
                base = fn
                break

    if base is None:
        return None
    elif base.startswith('/'):
        base = base[1:]
        return os.path.join(conf.paths.projectroot, base)
    else:
        return os.path.join(conf.paths.projectroot, conf.paths.builddata, base)

def image_tasks(conf, app):
    paths = conf.paths

    meta_file = get_images_metadata_file(conf)

    if 'images' not in conf.system.files.data:
        logger.info('no images to generate')
        return

    if isinstance(conf.system.files.data.images, list):
        images = conf.system.files.data.images
    else:
        images = [ conf.system.files.data.images ]

    for image in images:
        image['dir'] = paths.images
        source_base = os.path.join(paths.projectroot, image['dir'], image['name'])
        source_file = dot_concat(source_base, 'svg')
        rst_file = dot_concat(source_base, 'rst')
        image['conf'] = conf

        t = app.add('task')
        t.conf = conf
        t.job = generate_image_pages
        t.args = image
        t.description = "generating rst include file {0} for {1}".format(rst_file, source_file)
        t.target = rst_file
        t.dependency = meta_file
        logger.debug('adding task for image rst file: {0}'.format(rst_file))

        for output in image['output']:
            if 'tag' in output:
                tag = '-' + output['tag']
            else:
                tag = ''

            target_img = source_base + tag + '.png'

            inkscape_cmd = '{cmd} -z -d {dpi} -w {width} -y 0.0 -e >/dev/null {target} {source}'

            t = app.add('task')
            t.conf = conf
            t.job = _generate_images
            t.args = [ inkscape_cmd, output['dpi'], output['width'], target_img, source_file ]
            t.target = target_img
            t.dependency = [ meta_file, source_file ]
            t.description = 'generating image file {0} from {1}'.format(target_img, source_file)
            logger.debug('adding image creation job for {0}'.format(target_img))

def image_clean(conf, app):
    if 'images' not in conf.system.files.data:
        logger.info('no images to clean')
        return

    for image in conf.system.files.data.images:
        source_base = os.path.join(conf.paths.projectroot, conf.paths.images, image['name'])

        rm_rst = app.add('task')
        rm_rst.job = verbose_remove
        rm_rst.args = dot_concat(source_base, 'rst')

        for output in image['output']:
            rm_tag_image = app.add('task')
            rm_tag_image.job = verbose_remove
            if 'tag' in output:
                rm_tag_image.args = dot_concat(hyph_concat(source_base, output['tag']), 'png')
            else:
                rm_tag_image.args = dot_concat(source_base, 'png')



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import re
import os
import logging

logger = logging.getLogger('giza.content.post.json_output')

from giza.tools.command import command
from giza.tools.strings import dot_concat
from giza.tools.files import expand_tree, copy_if_needed
from giza.tools.transformation import munge_content

########## Process Sphinx Json Output ##########

def json_output(conf):
    list_file = os.path.join(conf.paths.branch_output, 'json-file-list')
    public_list_file = os.path.join(conf.paths.public_site_output,
                                    'json', '.file_list')

    cmd = 'rsync --recursive --times --delete --exclude="*pickle" --exclude=".buildinfo" --exclude="*fjson" {src} {dst}'

    json_dst = os.path.join(conf.paths.public_site_output, 'json')

    if not os.path.exists(json_dst):
        logger.debug('created directories for {0}'.format(json_dst))
        os.makedirs(json_dst)

    builder = 'json'
    if 'edition' in conf.project and conf.project.edition != conf.project.name:
        builder += '-' + conf.project.edition

    command(cmd.format(src=os.path.join(conf.paths.branch_output, builder) + '/',
                       dst=json_dst))

    copy_if_needed(list_file, public_list_file)
    logger.info('deployed json files to local staging.')

def json_output_tasks(conf, app):
    regexes = [
        (re.compile(r'<a class=\"headerlink\"'), '<a'),
        (re.compile(r'<[^>]*>'), ''),
        (re.compile(r'&#8220;'), '"'),
        (re.compile(r'&#8221;'), '"'),
        (re.compile(r'&#8216;'), "'"),
        (re.compile(r'&#8217;'), "'"),
        (re.compile(r'&#\d{4};'), ''),
        (re.compile(r'&nbsp;'), ''),
        (re.compile(r'&gt;'), '>'),
        (re.compile(r'&lt;'), '<')
    ]

    outputs = []
    for fn in expand_tree('source', 'txt'):
        # path = build/<branch>/json/<filename>

        if 'edition' in conf.project and conf.project.edition != conf.project.name:
            path = os.path.join(conf.paths.branch_output,
                                'json-' + conf.project.edition,
                                os.path.splitext(fn.split(os.path.sep, 1)[1])[0])

        else:
            path = os.path.join(conf.paths.branch_output,
                                'json', os.path.splitext(fn.split(os.path.sep, 1)[1])[0])



        fjson = dot_concat(path, 'fjson')
        json = dot_concat(path, 'json')

        task = app.add('task')
        task.target = json
        task.dependency = fjson
        task.job = process_json_file
        task.description = "processing json file".format(json)
        task.args = [fjson, json, regexes, conf]

        outputs.append(json)

    list_file = os.path.join(conf.paths.branch_output, 'json-file-list')

    list_task = app.add('task')
    list_task.target = list_file
    list_task.job = generate_list_file
    list_task.args = [outputs, list_file, conf]

    output = app.add('app')
    out_task = output.add('task')
    out_task.job = json_output
    out_task.args = [conf]
    out_task.description = 'transfer json output to public directory'

def process_json_file(input_fn, output_fn, regexes, conf=None):
    if  os.path.isfile(input_fn) is False:
        return False

    with open(input_fn, 'r') as f:
        document = f.read()

    doc = json.loads(document)

    if 'body' in doc:
        text = doc['body'].encode('ascii', 'ignore')
        text = munge_content(text, regexes)

        doc['text'] = ' '.join(text.split('\n')).strip()

    if 'title' in doc:
        title = doc['title'].encode('ascii', 'ignore')
        title = munge_content(title, regexes)

        doc['title'] = title

    url = [ conf.project.url, conf.project.basepath ]
    url.extend(input_fn.rsplit('.', 1)[0].split(os.path.sep)[3:])

    doc['url'] = '/'.join(url) + '/'

    with open(output_fn, 'w') as f:
        f.write(json.dumps(doc))

    return True

def generate_list_file(outputs, path, conf):
    dirname = os.path.dirname(path)

    url = '/'.join([ conf.project.url, conf.project.basepath, 'json' ])

    if not os.path.exists(dirname):
        os.mkdir(dirname)

    with open(path, 'w') as f:
        for fn in outputs:
            if os.path.isfile(fn) is True:
                f.write( '/'.join([ url, fn.split('/', 3)[3:][0]]) )
                f.write('\n')

    logger.info('rebuilt inventory of json output.')



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import TestCase

from giza.core.task import MapTask, Task
from giza.config.main import Configuration
from giza.config.runtime import RuntimeStateConfig

class BaseTaskSuite(object):
    def test_configuration_object_validation_rejection(self):
        for i in [None, 1, 'config', {'config': 1}]:
            t = self.Task()

            with self.assertRaises(TypeError):
                t.conf = i

    def test_configuration_object_validation_acceptance(self):
        for i in [self.c, Configuration(), RuntimeStateConfig()]:
            t = self.Task()
            t.conf = i

            self.assertIs(i, t.conf)

    def test_job_validation_rejection(self):
        for i in [None, True, False, 1, 'foo']:
            t = self.Task()

            with self.assertRaises(TypeError):
                t.job = i

    def test_job_validation_acceptance(self):
        for i in [Configuration, sum, TypeError, map]:
            t = self.Task()
            t.job = i

            self.assertIs(i, t.job)

    def test_args_type_setting_dict(self):
        self.assertIsNone(self.task._args)
        self.assertIsNone(self.task.args_type)

        self.task.args = { 'foo': 1 }
        self.assertEqual(self.task.args_type, 'kwargs')

    def test_args_type_setting_args(self):
        self.assertIsNone(self.task._args)
        self.assertIsNone(self.task.args_type)

        self.task.args = [1, 2, 3, 4]
        self.assertEqual(self.task.args_type, 'args')

    def test_args_type_setting_tuple(self):
        self.assertIsNone(self.task._args)
        self.assertIsNone(self.task.args_type)

        self.task.args = (1, 2, 3, 4)
        self.assertEqual(self.task.args_type, 'args')

    def test_args_type_setting_string(self):
        self.assertIsNone(self.task._args)
        self.assertIsNone(self.task.args_type)

        self.task.args = 'None'
        self.assertEqual(self.task.args_type, 'args')

    def test_rebuild_requirment_checker_base(self):
        self.assertIsNone(self.task.target)
        self.assertIsNone(self.task.dependency)
        self.assertFalse(self.c.runstate.force)

        self.assertTrue(self.task.needs_rebuild)

    def test_rebuild_requirment_checker_dep(self):
        self.task.target = True

        self.assertIsNotNone(self.task.target)
        self.assertIsNone(self.task.dependency)
        self.assertFalse(self.c.runstate.force)

        self.assertTrue(self.task.needs_rebuild)

    def test_rebuild_requirment_checker_force(self):
        self.task.target = True
        self.task.dependency = True
        self.c.runstate.force = True

        self.assertIsNotNone(self.task.target)
        self.assertIsNotNone(self.task.dependency)
        self.assertTrue(self.c.runstate.force)

        self.assertTrue(self.task.needs_rebuild)

class TestTask(BaseTaskSuite, TestCase):
    @classmethod
    def setUp(self):
        self.c = Configuration()
        self.c.runstate = RuntimeStateConfig()
        self.task = Task()
        self.Task = Task
        self.task.conf = self.c

    def test_running_tasks(self):
        def sum_func(iterable, start):
            print(iterable, start)
            return sum(iterable, start)

        for i in (((1,2,3), 0), ([1,2,3], 0), {'iterable': [1,2,3], 'start': 0}):
            t = self.Task()

            t.job = sum_func
            t.args = i

            self.assertEqual(t.run(), 6)

class TestMapTask(BaseTaskSuite, TestCase):
    @classmethod
    def setUp(self):
        self.c = Configuration()
        self.c.runstate = RuntimeStateConfig()
        self.task = MapTask()
        self.Task = MapTask
        self.task.conf = self.c

    def test_running_map_task(self):
        t = self.Task()
        t.iter = [ (i, i+10) for i in range(10) ]
        t.job = sum

        for a,b in zip(t.run(), [10, 12, 14, 16, 18, 20, 22, 24, 26, 28]):
            self.assertEqual(a,b)



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
import logging
from shutil import rmtree

logger = logging.getLogger('giza.content.source')

from giza.content.dependencies import dump_file_hashes
from giza.tools.command import command
from giza.tools.files import InvalidFile

def transfer_source(conf, sconf):
    target = os.path.join(conf.paths.projectroot, conf.paths.branch_source)

    if not os.path.exists(target):
        os.makedirs(target)
        logger.debug('created directory for sphinx build: {0}'.format(target))
    elif not os.path.isdir(target):
        msg = '"{0}" exists and is not a directory'.format(target)
        logger.error(msg)
        raise InvalidFile(msg)

    source_dir = os.path.join(conf.paths.projectroot, conf.paths.source)

    # we don't want rsync to delete directories that hold generated content in
    # the target so we can have more incremental
    exclusions = "--exclude=" + ' --exclude='.join([ os.path.join('includes', 'steps'),
                                                     os.path.join('includes', 'toc') ])

    command('rsync --times --checksum --recursive {2} --delete {0}/ {1}'.format(source_dir, target, exclusions))

    source_exclusion(conf, sconf)
    dump_file_hashes(conf)

    logger.info('prepared source for sphinx build in {0}'.format(target))

def source_tasks(conf, sconf, app):
    t = app.add('task')
    t.job = transfer_source
    t.args = [conf, sconf]
    t.description = 'transferring source to {0}'.format(conf.paths.branch_source)
    logger.info('adding task to migrate source to {0}'.format(conf.paths.branch_source))

def source_exclusion(conf, sconf):
    for fn in sconf.excluded:
        fqfn = os.path.join(conf.paths.projectroot, conf.paths.branch_source, fn[1:])
        if os.path.exists(fqfn):
            if os.path.isdir(fqfn):
                rmtree(fqfn)
            else:
                os.remove(fqfn)
                logger.debug('redacted {0}'.format(fqfn))

    logger.info('redacted {0} files'.format(len(sconf.excluded)))



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os.path

logger = logging.getLogger('giza.config.system')

from giza.config.base import RecursiveConfigurationBase, ConfigurationBase
from giza.config.sphinx_local import SphinxLocalConfig
from giza.config.manpage import ManpageConfig
from giza.config.pdfs import PdfConfig
from giza.config.intersphinx import IntersphinxConfig
from giza.config.translate import TranslateConfig
from giza.config.corpora import CorporaConfig
from giza.config.redirects import HtaccessData
from giza.tools.serialization import ingest_yaml_list

class SystemConfig(RecursiveConfigurationBase):
    @property
    def make(self):
        return self.state['make']

    @make.setter
    def make(self, value):
        self.state['make'] = SystemMakeConfig(value)

    @property
    def tools(self):
        return self.state['tools']

    @tools.setter
    def tools(self, value):
        self.state['tools'] = SystemToolsConfig(value)

    @property
    def files(self):
        if 'files' not in self.state:
            self.files = []

        return self.state['files']

    @files.setter
    def files(self, value):
        if 'files' not in self.state:
            self.state['files'] = SystemConfigFiles(self.conf)
            self.files.paths = value

    @property
    def conf_file(self):
        if 'conf_file' not in self.state:
            self.conf_file = None

        return self.state['conf_file']

    @conf_file.setter
    def conf_file(self, value):
        pass

    @property
    def branched(self):
        return self.conf.project.branched

    @branched.setter
    def branched(self, value):
        logger.warning('branched state is stored in system not project.')

    @property
    def dependency_cache(self):
        if 'dependency_cache' not in self.state:
            self.dependency_cache = None

        return self.state['dependency_cache']

    @dependency_cache.setter
    def dependency_cache(self, value):
        if value is not None:
            self.state['dependency_cache'] = value
        else:
            p = [ self.conf.paths.projectroot, self.conf.paths.branch_output ]
            if self.conf.project.edition is None:
                p.append('dependencies.json')
            else:
                p.append('dependencies-' + self.conf.project.edition + '.json')

            self.state['dependency_cache'] = os.path.sep.join(p)

class SystemToolsConfig(ConfigurationBase):
    @property
    def pinned(self):
        return self.state['pinned']

    @pinned.setter
    def pinned(self, value):
        if isinstance(value, bool):
            self.state['pinned'] = value
        else:
            raise TypeError

    @property
    def ref(self):
        return self.state['ref']

    @ref.setter
    def ref(self, value):
        if value in ('HEAD', 'master') or len(value) == 40:
            self.state['ref'] = value
        else:
            raise TypeError

class SystemMakeConfig(ConfigurationBase):
    @property
    def generated(self):
        return self.state['generated']

    @generated.setter
    def generated(self, value):
        if isinstance(value, list):
            self.state['generated'] = value
        else:
            raise TypeError

    @property
    def static(self):
        return self.state['static']

    @static.setter
    def static(self, value):
        if isinstance(value, list):
            self.state['static'] = value
        else:
            raise TypeError

class SystemConfigFiles(RecursiveConfigurationBase):
    def __init__(self, conf):
        super(SystemConfigFiles, self).__init__(None, conf)

    @property
    def paths(self):
        return self.state['paths']

    @paths.setter
    def paths(self, value):
        if isinstance(value, list):
            self.state['paths'] = value
        else:
            raise TypeError

    @property
    def data(self):
        if 'data' not in self.state:
            self.data = self.conf.system.files.paths

        return self.state['data']

    @data.setter
    def data(self, value):
        if 'data' not in self.state:
            self.state['data'] = SystemConfigData(value, self.conf)

class SystemConfigData(RecursiveConfigurationBase):
    ## There shouldn't be any setters in this class. All items in this class
    ## must exist in SystemConfigPaths() objects.

    _always_list_configs = ('manpages', 'pdfs', 'htaccess', 'push', 'images')

    def __init__(self, obj, conf):
        super(SystemConfigData, self).__init__(None, conf)
        for fn in self.conf.system.files.paths:
            if isinstance(fn, dict):
                attr_name = fn.keys()[0]
            else:
                attr_name = os.path.splitext(fn)[0]

            self._option_registry.append(attr_name)
            setattr(self, attr_name, fn)

    def __getattr__(self, key):
        try:
            return object.__getattribute__(self, key)
        except AttributeError as e:
            if key in self._option_registry:
                if isinstance(key, dict):
                    basename, fn = key.items()[0]
                else:
                    fn = key
                    basename = os.path.splitext(fn)[0]

                if isinstance(self.state[key], list):
                    pass
                elif isinstance(self.state[key], dict):
                    if len(self.state[key]) == 1:
                        basename, fns = self.state[key].items()[0]
                        if isinstance(fns, list):
                            for fn in fns:
                                self._load_file(basename, fn)
                        else:
                            self._load_file(basename, fn)
                else:
                    self._load_file(basename, self.state[key])

                if len(self.state[key]) == 1 and key not in self._always_list_configs:
                    return self.state[key][0]
                else:
                     return self.state[key]
            else:
                m = 'key "{0}" in system.data object does not exist'.format(key)
                if not key.startswith('__'):
                    logger.debug(m)
                raise AttributeError(m)

    def __contains__(self, value):
        return value in self._option_registry

    def ingest(self, file_list=None):
        # this needs to be a noop, because some of the base constructors call
        # it, but we want this class to load data lazily.
        pass

    def _load_file(self, basename, fn):
        full_path = self._resolve_config_path(fn)

        if basename not in self.state:
            if os.path.isfile(full_path):
                # TODO we should make this process lazy with a more custom getter/setter
                self.state[basename] = self._resolve_config_data(full_path, basename)
                logger.debug('set sub-config {0} with data from {0}'.format(basename, full_path))
            else:
                self.state[basename] = []
                logger.warning('{0} does not exist. continuing.'.format(full_path))
        else:
            if os.path.isfile(full_path):
                d = self._resolve_config_data(full_path, basename)
            else:  # if os.path.isfile(self.state[basename]):
                full_path = self.state[basename]# .items()[0][1]
                d = self._resolve_config_data(self._resolve_config_path(full_path), basename)
            self._set_config_data(basename, fn, d)

    def keys(self):
        return self._option_registry

    def _set_config_data(self, basename, fn, d):
        if isinstance(self.state[basename], list):
            if isinstance(d, list):
                self.state[basename].extend(d)
            else:
                self.state[basename].append(d)
        else:
            if isinstance(self.state[basename], (ConfigurationBase, dict)) and len(self.state[basename]) == 1:
                self.state[basename] = []
            elif self.state[basename] != fn:
                self.state[basename] = [self.state[basename]]
            else:
                self.state[basename] = []

            if isinstance(d, list):
                self.state[basename].extend(d)
            else:
                self.state[basename].append(d)

    def _resolve_config_path(self, fn):
        if isinstance(fn, dict):
            if len(fn) == 1:
                fn = fn.values()[0]
            else:
                logger.error("unsupported config file specification: " + str(fn))

        if os.path.exists(fn):
            full_path = fn
        elif os.path.exists(os.path.join(os.getcwd(), fn)):
            full_path = os.path.join(os.getcwd(), fn)
        elif fn.startswith('/'):
            full_path = os.path.join(self.conf.paths.projectroot, fn[1:])
        else:
            full_path = os.path.join(self.conf.paths.projectroot,
                                     self.conf.paths.builddata, fn)

        return full_path

    def _resolve_config_data(self, fn, basename):
        logger.debug('resolving config data from file ' + fn)
        if fn is None:
            return []
        else:
            data = ingest_yaml_list(fn)

            mapping = {
                'sphinx_local': SphinxLocalConfig,
                'sphinx-local': SphinxLocalConfig,
                'manpages': ManpageConfig,
                'pdfs': PdfConfig,
                'intersphinx': IntersphinxConfig,
                'corpora': CorporaConfig,
            }
            # recur_mapping for config objects that subclass RecursiveConfigurationBase
            recur_mapping = {
                'translate': TranslateConfig,
            }

            if basename in mapping:
                data = [ mapping[basename](doc) for doc in data ]
            elif basename in recur_mapping:
                data = [ recur_mapping[basename](doc, self.conf) for doc in data ]
            elif basename == 'htaccess':
                l = HtaccessData()
                l.conf = self.conf
                l.extend(data)
                data = l

            if len(data) == 1 and basename not in self._always_list_configs:
                return data[0]
            else:
                return data



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

logger = logging.getLogger('giza.hash')

from rstcloth.rstcloth import RstCloth

#################### Rendering ####################

def generate_hash_file(fn, conf):
    r = RstCloth()

    if os.path.exists(fn):
        with open(fn, 'r') as f:
            existing = f.read()
    else:
        existing = []

    commit = conf.git.commit
    r.directive('|commit| replace', '``{0}``'.format(commit))

    try:
        if r.data == existing[:-1]:
            logger.info('no new commit(s), not updating {0} ({1})'.format(fn, commit[:10]))
            return True
    except TypeError:
        logger.warning('problem generating {0}, continuing'.format(fn))
        with file(fn, 'a'):
            os.utime(fn, times)
    else:
        r.write(fn)
        logger.info('regenerated {0} with new commit hash: {1}'.format(fn, commit[:10]))

def generate_release_file(release_fn, conf):
    release_root = os.path.dirname(release_fn)
    if not os.path.exists(release_root):
        os.makedirs(release_root)

    with open(release_fn, 'w') as f:
        f.write(conf.git.commit)

    logger.info('generated "{0}" with current release hash.'.format(release_fn))

#################### Worker ####################

def hash_tasks(conf, app):
    hash_fn = os.path.join(conf.paths.projectroot,
                           conf.paths.includes,
                           'hash.rst')

    hash_task = app.add('task')
    hash_task.job = generate_hash_file
    hash_task.args = [hash_fn, conf]
    hash_task.target = hash_fn
    hash_task.description = 'creating hash file: {0}'.format(hash_fn)

    release_fn = os.path.join(conf.paths.projectroot,
                              conf.paths.public_site_output,
                              'release.txt')

    release_task = app.add('task')
    release_task.job = generate_release_file
    release_task.args = [release_fn, conf]
    release_task.target = release_fn
    release_task.description = "creating release filename: {0}".format(release_fn)



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
import os
import logging
import json

'''
This module is used for extracting the data received from experiments created
with build_model.py. It requires a directory structure similar to that created
by build_model.py. It saves the data in a data.csv file that can easily be
viewed in any spreadsheet program. It should be automatically after
build_model.py but can also be used on it's own.
'''
logger = logging.getLogger('giza.translate.model_results')


def grab_data(json_file, out):
    '''This function grabs data from the log and prints it to the outfile

    :param string json_file: json file from the build_model experiment to copy from
    :param file out: open data file to write to
    '''

    with open(json_file, "r") as f:
        d = json.load(f)

    score_list = re.split(", |/| = | \(|=|\)", d['BLEU'])
    BLEU_score = score_list[0]
    gram1 = score_list[1]
    gram2 = score_list[2]
    gram3 = score_list[3]
    gram4 = score_list[4]
    BP = score_list[6]
    ratio = score_list[8]
    hyp_len = score_list[10]
    ref_len = score_list[12]

    out.write(','.join([str(d['i']), str(d['max_phrase_length']), str(d['order']),
                        d['reordering_language'], d['reordering_directionality'],
                        d['score_options'], d['smoothing'], d['alignment'],
                        d['reordering_orientation'], d['reordering_modeltype'], BLEU_score, gram1,
                        gram2, gram3, gram4, BP, ratio, hyp_len, ref_len]))


def aggregate_model_data(project_path):
    '''This function goes through the different log files and writes the
    data to the outfile

    :param string project_path: path to the model as specified in the config file
    '''
    with open("{0}/data.csv".format(project_path), "w", 1) as out:
        out.write("i,max phrase length,order,reordering language,reordering directionality,score options,smoothing,alignment,reordering orientation,reordering modeltype,BLEU Score,1-gram precision,2-gram precision,3-gram precision,4-gram precision,BP,ratio,hyp len,ref len\n")
        i = 0
        while True:
            json_path = os.path.join(project_path, str(i), str(i)) + ".json"
            if os.path.isfile(json_path) is False:
                break
            grab_data(json_path, out)
            i += 1



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division
import os
import logging
import re
import math

import yaml
import polib

from giza.translate.utils import get_file_list

''''
This module contains functions that involve creating corpora. It has one
function that creates the appropriate train, tune, and test hybrid corpora from
smaller corpora. It has other functions that create corpora from already
translated po files as well as from multi-lingual dictionaries
'''

logger = logging.getLogger('giza.translate.corpora')


def append_corpus(percentage, num_copies, out_fn, in_fn, start, final=False):
    '''This function appends the correct amount of the corpus to the basefile,
    finishing up the file when necessary so no data goes to waste

    :param int percentage: percentage of the file going into the corpus
    :param float num_copies: number of copies of the file going into the corpus
    :param string out_fn: the name of the base file to append the corpus to
    :param string in_fn: the name of the new file to take the data from
    :param int start: the line to start copying from in the new file
    :param boolean final: if it's the final section of the file. If True it
         makes sure to use all of the way to the end of the file

    :returns: the last line it copied until
    '''
    with open(in_fn, 'r') as f:
        new_content = f.readlines()

    with open(out_fn, 'a+') as f:
        tot = int(len(new_content) * percentage / 100)
        for i in range(int(math.floor(num_copies))):
            if final is False:
                f.writelines(new_content[start:start + tot])
            else:
                f.writelines(new_content[start:])
            # if there's no new line at the end this adds one
            f.seek(-1, os.SEEK_END)
            if f.read(1) != '\n':
                f.write('\n')

        # if we have a fractional number of copies then we take care of the rest
        f.writelines(new_content[start:start + int(tot*(num_copies - math.floor(num_copies)))])

        # if there's no new line at the end this adds one
        f.seek(-1, os.SEEK_END)
        if f.read(1) != '\n':
            f.write('\n')

    return start + tot


def get_total_length(conf, corpus_type):
    '''This function finds the ideal total length of the corpus
    It finds the minimum length where each corpus section is used in full

    :param config conf: corpora configuration object
    :param string corpus_type: either train, tune, or test

    :returns: total length of the corpus
    '''
    tot_length = 0
    for file_name, source in conf.sources.items():
        if source.state['percent_of_' + corpus_type] > 0:
            temp_length = source.length*source.state['percent_' + corpus_type] / source.state['percent_of_' + corpus_type]
            if temp_length > tot_length:
                tot_length = temp_length
    return tot_length


def create_hybrid_corpora(conf):
    '''This function takes the configuration file and runs through the files,
    appending them appropriately. It first verifies that all of the percentages
    add up and then it figures out how much of each file should go into each
    corpus and appends them. The config file should be similar to corpora.yaml.
    It will copy the config file to the directory with the corpora so that you
    have a record, but the copy won't be exact. It creates both language
    corpora at the same time in parallel.

    :param config conf: corpora configuration object
    '''

    os.makedirs(conf.container_path)
    with open(os.path.join(conf.container_path, "corpora.yaml"), 'w') as f:
        yaml.dump(conf.dict(), f, default_flow_style=False)

    # append files appropriately
    for corpus_type in ('train', 'tune', 'test'):
        source_outfile = os.path.join(conf.container_path, "{0}.{1}-{2}.{1}".format(corpus_type,
                                                                                    conf.source_language,
                                                                                    conf.target_language))
        target_outfile = os.path.join(conf.container_path, "{0}.{1}-{2}.{2}".format(corpus_type,
                                                                                    conf.source_language,
                                                                                    conf.target_language))
        open(source_outfile, 'w').close()
        open(target_outfile, 'w').close()

        tot_length = get_total_length(conf, corpus_type)

        for fn, source in conf.sources.items():
            logger.info("Processing " + fn + " for " + corpus_type)

            # finds how many copies of this file will make it the correct percentage of the full corpus
            if source.state['percent_' + corpus_type] == 0 or source.state['percent_of_' + corpus_type] == 0:
                continue
            num_copies = tot_length * source.state['percent_of_' + corpus_type] / source.length / source.state['percent_' + corpus_type]
            final = False
            if corpus_type is 'test':
                final = True

            # appends the section of the file to the corpora
            end = append_corpus(source.state['percent_' + corpus_type],
                                num_copies,
                                source_outfile,
                                source.source_file_path,
                                source.end, final)
            append_corpus(source.state['percent_' + corpus_type],
                          num_copies,
                          target_outfile,
                          source.target_file_path,
                          source.end,
                          final)
            source.end = end


def write_from_po_file(source_doc, target_doc, po_file_name):
    '''This function writes two files in english and spanish from a po file's
    translated entries

    :param string source_doc: Name of file to put source lanaguge text in.
    :param string target_doc: Name of file to put target lanaguge text in.
    :param string po_file_name: Path to po file to parse
    '''
    logger.info("processing "+po_file_name)
    po = polib.pofile(po_file_name)
    for entry in po.translated_entries():
        source_doc.write(entry.msgid.encode('utf-8')+'\n')
        target_doc.write(entry.msgstr.encode('utf-8')+'\n')


def create_corpus_from_po(po_path, source_doc_fn, target_doc_fn):
    '''This function opens up the output files and then goes through the files
    in the file list and writes them all to two corpus files.

    :param string po_path: Path to po file or directory of po files
    :param string source_doc_fn: Name of file to put source lanaguge text in.
    :param string target_doc_fn: Name of file to put target lanaguge text in.
    '''

    # path is a directory now
    logger.info("walking path "+po_path)
    with open(source_doc_fn, "w", 1) as source_doc:
        with open(target_doc_fn, "w", 1) as target_doc:
            file_list = get_file_list(po_path, ["po", "pot"])
            for fn in file_list:
                write_from_po_file(source_doc, target_doc, fn)


def create_corpus_from_dictionary(dict_fn, source_fn, target_fn):
    '''This function splits a dictionary from http://www.dicts.info/uddl.php
    and turns it into a parallel corpus.

    :param string dict_fn: path to dictionary file
    :param string source_fn: path to file to write source text to
    :param string target_fn: path to file to write target text to
    '''

    with open(dict_fn, "r") as dict_f:
        with open(source_fn, "w", 1) as source_f:
            with open(target_fn, "w", 1) as target_f:
                for line in dict_f:
                    if line[0] == '#':
                        continue
                    halves = re.split(r'\t+', line)
                    if len(halves) < 2:
                        continue
                    source_words = re.split(r';', halves[0])
                    target_words = re.split(r';', halves[1])
                    for target_word in target_words:
                        target_word = target_word.strip()
                        for source_word in source_words:
                            source_word = source_word.strip()
                            source_f.write(source_word)
                            source_f.write("\n")
                            target_f.write(target_word)
                            target_f.write("\n")



import logging
import argh

logger = logging.getLogger('giza.operations.deploy')

from giza.config.helper import fetch_config
from giza.core.app import BuildApp
from giza.deploy import Deploy
from giza.operations.sphinx import sphinx_publication
from giza.tools.command import verbose_command
from giza.tools.serialization import ingest_yaml_list, dict_from_list

@argh.arg('--target', '-t', nargs='*', dest='push_targets')
@argh.arg('--dry-run', '-d', action='store_true', dest='dry_run')
@argh.named('deploy')
def main(args):
    c = fetch_config(args)
    app = BuildApp(c)

    deploy_worker(c, app)

@argh.arg('--deploy', '-d', nargs='*', dest='push_targets')
@argh.arg('--edition', '-e', nargs='*', dest='editions_to_build')
@argh.arg('--language', '-l', nargs='*',dest='languages_to_build')
@argh.arg('--builder', '-b', nargs='*', default='html')
@argh.arg('--serial_sphinx', action='store_true', default=False)
@argh.named('push')
def publish_and_deploy(args):
    c = fetch_config(args)
    app = BuildApp(c)

    sphinx_ret = sphinx_publication(c, args, app)
    if sphinx_ret == 0 or c.runstate.force is True:
        deploy_worker(c, app)
    else:
        logger.warning(sphinx_ret + ' sphinx build(s) failed, and build not forced. not deploying.')

def deploy_worker(c, app):
    pconf = c.system.files.data.push
    pconf = dict_from_list('target', pconf)

    cmds = []
    for target in c.runstate.push_targets:
        d = Deploy(c)

        target_pconf = pconf[target]

        if target_pconf['env'] == 'publication':
            target_pconf['env'] = 'production'

        d.load(target_pconf)

        map_task = app.add('map')
        map_task.iter = d.deploy_commands()

        map_task.job = verbose_command

        cmds.extend(d.deploy_commands())

    if c.runstate.dry_run is True:
        for i in cmds:
            logger.info('dry run: {0}'.format(' '.join(i)))
    else:
        app.run()

    logger.info('completed deploy for: {0}'.format(' '.join(c.runstate.push_targets)))



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import itertools
import json
import logging
import os

import argh
import yaml

logger = logging.getLogger('giza.scrumpy')

from giza.config.jeerah import fetch_config, JeerahRuntimeStateConfig
from giza.config.helper import dump_skel
from giza.config.credentials import setup_credentials
from giza.cmdline import get_base_parser
from giza.core.app import BuildApp
from giza.jeerah.client import JeerahClient
from giza.jeerah.query import strip_name

import giza.jeerah.progress
import giza.jeerah.triage

#################### helpers #####################

def pprint(obj):
    print(json.dumps(obj, indent=3))

#################### scrumpy commands #####################

def config(args):
    conf = fetch_config(args)

    dynamics = [ conf.site, conf.runstate, conf.buckets, conf.sprints,
                 conf.reporting, conf.modification ]

    pprint(json.dumps(conf.dict(), indent=3))

@argh.arg('--sprint')
def progress(args):
    conf = fetch_config(args)
    app = BuildApp(conf)
    app.pool = 'thread'

    j = JeerahClient(conf)
    j.connect()

    query_data = giza.jeerah.progress.query(j, app, conf)

    pprint(giza.jeerah.progress.report(query_data, conf))

def triage(args):
    conf = fetch_config(args)
    app = BuildApp(conf)
    app.pool = 'thread'

    j = JeerahClient(conf)
    j.connect()

    query_data = giza.jeerah.triage.query(j, app, conf)

    pprint(giza.jeerah.triage.report(query_data, conf))

#################### Jira Modification Tasks ####################

@argh.named('create-versions')
@argh.arg('--sprint')
@argh.arg('--project')
def make_versions(args):
    conf = fetch_config(args)

    j = JeerahClient(conf)
    j.connect()

    current_versions = [strip_name(v.name) for v in j.versions(conf.runstate.project)]

    created = []
    for v in getattr(conf.sprints, conf.runstate.sprint):
        v = strip_name(v)
        if v not in current_versions:
            j.create_version(conf.runstate.project, v, release=False)
            created.append(v)
            logger.info('creating new version {0} in project {1}'.format(v, conf.runstate.project))
        else:
            logger.info('version {0} already exists in project {1}'.format(v, conf.runstate.project))

    pprint({ 'created': created, 'project': conf.runstate.project })

@argh.named('mirror-versions')
def mirror_version(args):
    results = { 'created': {}, 'targets': {} }

    conf = fetch_config(args)

    j = JeerahClient(conf)
    j.connect()

    source_project = conf.modification.mirroring.source

    results['sources'] = [ strip_name(v.name)
                           for v in j.versions(source_project)
                         ]

    for target_project in conf.modification.mirroring.target:
        results['created'][target_project] = []
        results['targets'][target_project] = [ v.name
                                               for v in j.versions(target_project)
                                              ]
        for i in results['sources']:
            if i not in results['targets'][target_project]:
                j.create_version(target_project, i, '', False)
                logger.info('created version named "{0}" in {1} project'.format(i, target_project))
                results['created'][target_project].append(i)
            else:
                logger.info('project {0} exists. passing.'.format(i))

    pprint(results)

def release(args):
    results = { }

    conf = fetch_config(args)

    j = JeerahClient(conf)
    j.connect()

    for project in conf.site.projects:
        results[project] = []
        current = j.versions(project)

        for version in conf.sprints.current:
            for v in current:
                if v.name in version:
                    logger.debug('archiving {0} in project {1}'.format(v.name, project))
                    j.archive_version(v)
                    j.release_version(v)
                    results[project].append(v.name)
                else:
                    logger.debug('{0} is not eligible for release'.format(v.name))

    pprint({ 'released': results, 'code': 200 })


@argh.arg('--path', dest='user_conf_path', default='.scrumpy.yaml')
def setup(args):
    skel = {
        'site': { 'credentials': "~/.giza-credentials.yaml",
                  'projects': ['DOCS', 'TOOLS', 'INTERNAL'],
                  'url': "https://jira.example.net/" },
        'sprints': { 'previous': [ 'sprint-0', 'release-0' ],
                     'current': [ 'spirnt-1', 'release-1'],
                     'next': [ 'spirnt-2', 'release-2'],
                     'future': [ 'spirnt-3', 'release-3'],
                     'archived': [
                         [ 'legacy-0', 'legacy-release-0'],
                         [ 'legacy-1', 'legacy-release-1']
                      ],
        },
        'buckets': {
            'next': 'docs-next',
            'planning': 'docs-planning',
            'triage': 'docs-triage'
        },
        'reporting': {
            'units': 'days',
            'format': 'json'
        },
        'modification': {
            'mirroring': {
                'source': 'DOCS',
                'target': 'INTERNAL'
            }
        }
    }

    dump_skel(skel, args)

@argh.named('setup-credentials')
def setup_credential_file(args):
    setup_credentials(args)

#################### scrumpy entry point #####################

def main():
    parser = get_base_parser()

    commands = [setup, setup_credential_file, config, progress, triage,
                make_versions, mirror_version, release]

    argh.add_commands(parser, commands)

    args = JeerahRuntimeStateConfig()

    if args.level == 'info':
        args.level = 'warning'

    argh.dispatch(parser, namespace=args)



import logging
import os
import re
import subprocess

logger = logging.getLogger('giza.content.post.latex')

from giza.content.helper import edition_check
from giza.tools.command import command
from giza.tools.serialization import ingest_yaml_list
from giza.tools.strings import hyph_concat
from giza.tools.transformation import munge_page
from giza.tools.files import (create_link, copy_if_needed,
                        decode_lines_from_file, encode_lines_to_file)

#################### PDFs from Latex Produced by Sphinx  ####################

def _clean_sphinx_latex(fn, regexes):
    munge_page(fn, regexes, tag='pdf')

def _render_tex_into_pdf(fn, path):
    pdflatex = 'TEXINPUTS=".:{0}:" pdflatex --interaction batchmode --output-directory {0} {1}'.format(path, fn)

    base_fn = os.path.basename(fn)
    cmds = [ pdflatex,
             "makeindex -s {0}/python.ist {0}/{1}.idx ".format(path, base_fn[:-4]),
             pdflatex,
             pdflatex ]

    for idx, cmd in enumerate(cmds):
        r = command(command=cmd, ignore=True)

        if r.succeeded is True:
            logger.info('pdf completed rendering stage {0} of {1} successfully.'.format(idx, len(cmds)))
        else:
            if idx <= 1:
                logger.warning('pdf build encountered error early on {0}, continuing cautiously.'.format(base_fn))
                continue
            else:
                logger.error('pdf build encountered error running pdflatex, investigate on {0}. terminating'.format(base_fn))
                return False

def pdf_tasks(sconf, conf, app):
    target = sconf.builder
    if 'pdfs' not in conf.system.files.data:
        return

    app.pool = 'thread'

    tex_regexes = [ ( re.compile(r'(index|bfcode)\{(.*)--(.*)\}'),
                      r'\1\{\2-\{-\}\3\}'),
                    ( re.compile(r'\\PYGZsq{}'), "'"),
                    ( re.compile(r'\\code\{/(?!.*{}/|etc|usr|data|var|srv)'),
                      r'\code{' + conf.project.url + r'/' + conf.project.tag) ]

    clean_app = app.add('app')
    cache_app = app.add('app')
    render_app = app.add('app')
    migrate_app = app.add('app')
    link_app = app.add('app')

    for i in conf.system.files.data.pdfs:
        if edition_check(i, conf) is False:
            continue

        i = i.dict()
        tagged_name = i['output'][:-4] + '-' + i['tag']
        deploy_fn = tagged_name + '-' + conf.git.branches.current + '.pdf'
        link_name = deploy_fn.replace('-' + conf.git.branches.current, '')

        if 'edition' in conf.project and conf.project.edition != conf.project.name:
            if 'edition' in i and conf.project.edition != i['edition']:
                continue
            latex_dir = os.path.join(conf.paths.branch_output, hyph_concat(target, conf.project.edition))
        else:
            latex_dir = os.path.join(conf.paths.branch_output, target)

        deploy_path = conf.paths.public_site_output

        i['source'] = os.path.join(latex_dir, i['output'])
        i['processed'] = os.path.join(latex_dir, tagged_name + '.tex')
        i['pdf'] = os.path.join(latex_dir, tagged_name + '.pdf')
        i['deployed'] = os.path.join(deploy_path, deploy_fn)
        i['link'] = os.path.join(deploy_path, link_name)
        i['path'] = latex_dir

        clean_task = clean_app.add('task')
        clean_task.target = i['source']
        clean_task.job = _clean_sphinx_latex
        clean_task.args = (i['source'], tex_regexes)

        cache_task = cache_app.add('task')
        cache_task.dependency = i['source']
        cache_task.target = i['processed']
        cache_task.job = copy_if_needed
        cache_task.args = (i['source'], i['processed'], 'pdf')

        render_task = render_app.add('task')
        render_task.dependency = i['processed']
        render_task.target = i['pdf']
        render_task.job = _render_tex_into_pdf
        render_task.args = (i['processed'], i['path'])

        migrate_task = migrate_app.add('task')
        migrate_task.dependency = i['pdf']
        migrate_task.target = i['deployed']
        migrate_task.job = copy_if_needed
        migrate_task.args = (i['pdf'], i['deployed'], 'pdf')

        if i['link'] != i['deployed']:
            link_task = link_app.add('task')
            link_task.dependency = i['deployed']
            link_task.target = i['link']
            link_task.job = create_link
            link_task.args = (deploy_fn, i['link'])



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
import datetime

def concat(*args):
    return ''.join(args)

def dot_concat(*args):
    return '.'.join(args)

def hyph_concat(*args):
    return '-'.join(args)

def slash_concat(*args):
    return '/'.join(args)

def path_concat(*args):
    return os.path.sep.join(args)

def timestamp(form='filename'):
    if form == 'filename':
        return datetime.datetime.now().strftime("%Y-%m-%d.%H-%M")
    else:
        return datetime.datetime.now().strftime("%Y-%m-%d, %H:%M %p")



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import os
import logging

logger = logging.getLogger('giza.content.manpages')

from giza.tools.files import expand_tree

#################### Manpage Processing ####################

def manpage_url(regex_obj, input_file):
    with open(input_file, 'r') as f:
        manpage = f.read()

    if isinstance(regex_obj, list):
        for regex, subst in regex_obj:
            manpage = regex.sub(subst, manpage)
    else:
        manpage = regex_obj[0].sub(regex_obj[1], manpage)

    with open(input_file, 'w') as f:
        f.write(manpage)

    logger.info("fixed urls in {0}".format(input_file))

def manpage_url_tasks(builder, conf, app):
    project_source = os.path.join(conf.paths.projectroot,
                                  conf.paths.source)

    top_level_items = set()
    for fs_obj in os.listdir(project_source):
        if fs_obj.startswith('.static') or fs_obj == 'index.txt':
            continue
        if os.path.isdir(os.path.join(project_source, fs_obj)):
            top_level_items.add(fs_obj)
        if fs_obj.endswith('.txt'):
            top_level_items.add(fs_obj[:-4])

    top_level_items = '/'+ r'[^\s]*|/'.join(top_level_items) + r'[^\s]*'

    re_string = r'(\\fB({0})\\fP)'.format(top_level_items).replace(r'-', r'\-')
    subst = conf.project.url + '/' + conf.project.tag + r'\2'

    regex_obj = (re.compile(re_string), subst)

    for manpage in expand_tree(os.path.join(conf.paths.projectroot,
                                            conf.paths.output,
                                            conf.git.branches.current,
                                            builder), ['1', '5']):
        task = app.add('task')
        task.target = manpage
        task.job = manpage_url
        task.args = [regex_obj, manpage]
        task.description = 'processing urls in manpage file: {0}'.format(manpage)



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
from copy import deepcopy

logger = logging.getLogger('giza.content.options')

from jinja2 import Template
from rstcloth.rstcloth import RstCloth

from giza.tools.files import expand_tree, rm_rf
from giza.tools.serialization import ingest_yaml_list

#################### Rendering ####################

class Options(object):
    def __init__(self, fn=None):
        self.data = list()
        self.cache = dict()
        self.source_files = list()
        self.unresolved = list()

        if fn is not None:
            self.source_dirname = os.path.dirname(os.path.abspath(fn))
            logger.debug("reading options from {0}".format(fn))
            self.ingest(fn)
        else:
            self.source_dirname = fn

    def ingest(self, fn):
        if self.source_dirname is None:
            self.source_dirname = os.path.dirname(os.path.abspath(fn))

        input_sources = ingest_yaml_list(os.path.join(self.source_dirname, os.path.basename(fn)))
        self.cache[fn] = dict()

        self.source_files.append(input_sources)

        for option in input_sources:
            opt = Option(option)

            self.cache_option(opt, fn)

        self.resolve(fn)

    def cache_option(self, opt, fn):
        if opt.program not in self.cache[fn]:
            self.cache[fn][opt.program] = dict()

        self.cache[fn][opt.program][opt.name] = opt

        if opt.inherited is True:
            self.unresolved.append(opt)
        else:
            self.data.append(opt)

    def resolve(self, fn):
        for opt in self.unresolved:
            if opt.inherited is True:
                if opt.source['file'] == fn:
                    continue

                if opt.source['file'] in self.cache:
                    base_opt = self.resolve_inherited(opt.source)
                else:
                    self.ingest(opt.source['file'])
                    base_opt = self.resolve_inherited(opt.source)

                base_opt.doc.update(opt.doc)
                if 'inherit' in base_opt.doc:
                    del base_opt.doc['inherit']

                opt = Option(base_opt.doc)
                logmsg = 'cached option while rendering {0}, program "{1}" name "{2}"'
                logger.debug(logmsg.format(self.source_dirname, opt.program, opt.name))

                self.cache_option(opt, fn)

        self.unresolved = list()

    def resolve_inherited(self, spec):
        return deepcopy(self.cache[spec['file']][spec['program']][spec['name']])

    def iterator(self):
        for item in self.data:
            if not item.program.startswith('_'):
                yield item

optional_source_fields = ['default', 'type', 'pre', 'description', 'post', 'directive']

class Option(object):
    def __init__(self, doc):
        self.doc = doc
        self._directive = None

        self.name = doc['name']
        self.program = doc['program']

        for k in optional_source_fields:
            if k in doc:
                setattr(self, k, doc[k])
            else:
                setattr(self, k, None)

        if 'inherit' in doc:
            self.inherited = True
            self.source = doc['inherit']
        else:
            self.inherited = False
            self.source = {}

        if 'args' in doc:
            if doc['args'] is None or doc['args'] == '':
                pass
            else:
                self.arguments = doc['args']

        if 'aliases' in doc:
            if doc['aliases'] is None or doc['aliases'] == '':
                pass
            else:
                if isinstance(doc['aliases'], list):
                    self.aliases = doc['aliases']
                else:
                    self.aliases = [ doc['aliases'] ]

        self.replacement = dict()
        if 'replacement' in doc:
            if isinstance(doc['replacement'], dict):
                self.replacement = doc['replacement']

        # add auto-populated replacements here
        self.add_replacements()
        logmsg = 'created option representation for the {0} option for {1}'
        logger.debug(logmsg.format(self.name, self.program))

    @property
    def directive(self):
        if self._directive is None:
            return 'describe'
        else:
            return self._directive

    @directive.setter
    def directive(self, value):
        self._directive = value

    def add_replacements(self):
        if not self.program.startswith('_'):
            if 'program' not in self.replacement:
                self.replacement['program'] = ':program:`{0}`'.format(self.program)
            if 'role' not in self.replacement:
                if self.directive == 'describe':
                    self.replacement['role'] = "``{0}``".format(self.name)
                elif self.directive == 'option':
                    self.replacement['role'] = ":{0}:`--{1}`".format(self.directive, self.name)
                else:
                    self.replacement['role'] = ":{0}:`{1}`".format(self.directive, self.name)

    def replace(self):
        attempts = range(10)

        if self.description is not None:
            for i in attempts:
                template = Template(self.description)
                self.description = template.render(**self.replacement)

                if "{{" not in self.description:
                    break

        if self.pre is not None:
            for i in attempts:
                template = Template(self.pre)
                self.pre = template.render(**self.replacement)

                if "{{" not in self.pre:
                    break

        if self.post is not None:
            for i in attempts:
                template = Template(self.post)
                self.post = template.render(**self.replacement)

                if "{{" not in self.post:
                    break

        if self.default is not None and not (isinstance(self.default, bool) or
                                             isinstance(self.default, int)):
            for i in attempts:
                template = Template(self.default)
                self.efault = template.render(**self.replacement)

                if "{{" not in self.default:
                    break

class OptionRendered(object):
    def __init__(self, option):
        if not isinstance(option, Option):
            raise TypeError
        else:
            self.option = option

        self.rst = RstCloth()

    def resolve_option_name(self):
        if self.option.directive == 'option':
            if self.option.name.startswith('<'):
                prefix = ''
            else:
                prefix = '--'

            if hasattr(self.option, 'aliases'):
                if hasattr(self.option, 'arguments'):
                    return '{0}{1} {2}, {3}'.format(prefix, self.option.name,
                                                    self.option.arguments,
                                                    '{0}, '.format(self.option.arguments).join(self.option.aliases))
                else:
                    return '{0}{1}, {2}'.format(prefix, self.option.name,
                                                ', '.join(self.option.aliases))

            else:
                if hasattr(self.option, 'arguments'):
                    return '{0}{1} {2}'.format(prefix, self.option.name,
                                               self.option.arguments)
                else:
                    return '{0}{1}'.format(prefix, self.option.name)
        else:
            return self.option.name

    def resolve_output_path(self, path):
        name_parts = self.option.name.split(',')

        if len(name_parts) > 1:
            clensed_name = name_parts[0]
        else:
            clensed_name = self.option.name

        fn = '-'.join([ self.option.directive, self.option.program, clensed_name ]) + '.rst'
        return os.path.join(path, fn)

    def render(self, path):
        self.option.replace()

        self.rst.directive(self.option.directive, self.resolve_option_name())
        self.rst.newline()

        if self.option.type is not None:
            self.rst.content('*Type*: {0}'.format(self.option.type), indent=3)
            self.rst.newline()

        if self.option.default is not None:
            self.rst.content('*Default*: {0}'.format(self.option.default), indent=3)
            self.rst.newline()

        if self.option.pre is not None:
            self.rst.content(self.option.pre.split('\n'), indent=3, wrap=False)
            self.rst.newline()

        if self.option.description is not None:
            self.rst.content(self.option.description.split('\n'), indent=3, wrap=False)
            self.rst.newline()

        if self.option.post is not None:
            self.rst.content(self.option.post.split('\n'), indent=3, wrap=False)
            self.rst.newline()

        output_file = self.resolve_output_path(path)

        self.rst.write(output_file)

        logger.debug('wrote option to file {0}'.format(output_file))


#################### Source Iterators ####################1

def get_option_path(conf):
    return os.path.join(get_base_path(conf), 'option')

def get_base_path(conf):
    return os.path.join(conf.paths.projectroot, conf.paths.includes)

def option_sources(conf):
    output_path = get_option_path(conf)

    for fn in expand_tree(get_base_path(conf), 'yaml'):
        if fn.startswith(output_path):
            yield fn

#################### Worker and Integration ####################1

def render_option_page(opt, path):
    renderer = OptionRendered(opt)
    renderer.render(path)

def option_tasks(conf, app):
    base_path = get_base_path(conf)

    options = Options()
    options.source_dirname = base_path
    output_path = get_option_path(conf)

    for fn in option_sources(conf):
        options.ingest(fn)

    if not os.path.exists(output_path):
        os.makedirs(output_path)

    for opt in options.iterator():
        t = app.add('task')
        t.job = render_option_page
        t.args = [ opt, output_path ]
        t.description = 'generating option "{0}" for "{1}"'.format(opt.name, opt.program)

def option_clean(conf):
    rm_rf(get_option_path(conf))
    logger.info('removed all option output')



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
:mod:`~giza.inheritance` is a library for managing content reuse using an
inheritance-based model for structured content. In systems that use this code,
writers define content in a structured form, where content units can "inherit"
from other content units and optionally override some portions of the inherited
units.
"""

import logging
import os.path

logger = logging.getLogger('giza.content.examples.inheritance')

from giza.config.base import RecursiveConfigurationBase, ConfigurationBase
from giza.tools.serialization import ingest_yaml_list

class InheritableContentError(Exception):
    """
    Exception used by inheritance code to indicate a problem resolving
    inherited content.
    """

    pass

class InheritableContentBase(RecursiveConfigurationBase):
    """
    Base data object that represents a single unit of content. Typically
    sub-classed.
    """

    _option_registry = ['pre', 'post', 'final', 'ref', 'content', 'edition']

    @property
    def title(self):
        return self.state['title']

    @title.setter
    def title(self, value):
        if isinstance(value, basestring):
            self.state['title'] = TitleData({'text': value})
        elif isinstance(value, TitleData):
            self.state['title'] = value
        elif isinstance(value, dict):
            self.state['title'] = TitleData(value)
        else:
            print value, type(value)
            raise TypeError

    @property
    def source(self):
        if 'source' in self.state:
            return self.state['source']
        else:
            return None

    @source.setter
    def source(self, value):
        self.state['source'] = InheritanceReference(value, self.conf)

    inherit = source
    def is_resolved(self):
        if self.source is None:
            return True
        else:
            try:
                return self.source.resolved
            except AttributeError:
                return False

    def _is_resolveable(self, data):
        try:
            return not self.is_resolved() and self.source.file in data
        except AttributeError:
            logger.warning(str(data) + ' is not resolvable')
            return False


    def resolve(self, data):
        if self._is_resolveable(data):
            try:
                base = data.fetch(self.source.file, self.source.ref)
                base.resolve(data)

                base.state.update(self.state)
                self.state.update(base.state)
                self.source.resolved = True
                return True
            except InheritableContentError as e:
                logger.error(e)

        if self.source is not None and not self.source.is_resolved():
            m = 'cannot find {0} and ref {1} do not  exist'.format(self.source.file, self.source.ref)
            logger.error(m)
            raise InheritableContentError(m)

class InheritanceReference(RecursiveConfigurationBase):
    """
    Represents a single reference to another unit of content. The
    setter method for the :meth:`~giza.inheritance.InheritanceReference.file`
    attribute, returns an error if it cannot discover a specified value.
    """

    _option_registry = ['ref']

    @property
    def resolved(self):
        if 'resolved' not in self.state:
            return False
        else:
            return self.state['resolved']

    @resolved.setter
    def resolved(self, value):
        if isinstance(value, bool):
            self.state['resolved'] = value
        else:
            raise TypeError('{0} is not boolean'.format(value))

    @property
    def file(self):
        return self.state['file']

    @file.setter
    def file(self, value):
        fns = [ value,
                os.path.join(self.conf.paths.projectroot, value),
                os.path.join(self.conf.paths.projectroot,
                             self.conf.paths.includes, value) ]

        try:
            fns.append(os.path.join(self.conf.paths.projectroot,
                                    self.conf.paths.source, value))
        except KeyError:
            pass


        for fn in fns:
            if os.path.exists(fn):
                self.state['file'] = fn
                break

        if 'file' not in self.state:
            raise TypeError('file named {0} does not exist'.format(value))


##############################

class DataContentBase(RecursiveConfigurationBase):
    """
    Represents a group of units ingested from a single file.
    """

    content_class = InheritableContentBase

    def __init__(self, src, data, conf):
        self._state = { 'content': { } }
        self._content = self._state['content']
        self._conf = None
        self.conf = conf
        self.data = data
        self.ingest(src)

    def __contains__(self, value):
        if value in self.content:
            return True
        else:
            return False

    @property
    def content(self):
        return self.state['content']

    @content.setter
    def content(self, value):
        logger.warning('cannot set content record directly')

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        if isinstance(value, DataCache):
            self._data = value
        else:
            logger.warning('cannot use invalid data cache instance.')

    def ingest(self, src):
        if not isinstance(src, list) and os.path.exists(src):
            src = ingest_yaml_list(src)

        for doc in src:
            self.add(doc)

    def add(self, doc):
        if doc['ref'] not in self.content:
            if isinstance(doc, self.content_class):
                self.content[doc['ref']] = doc
            else:
                try:
                    self.content[doc['ref']] = self.content_class(doc)
                except TypeError:
                    self.content[doc['ref']] = self.content_class(doc, self.conf)
        else:
            m = 'content named {0} already exists'.format(doc['ref'])
            logger.error(m)
            raise InheritableContentError(m)

    def fetch(self, ref):
        if ref in self.content:
            content = self.content[ref]

            if not content.is_resolved():
                content.resolve(self.data)

            return content
        else:
            m = 'content with ref "{0}" not found'.format(ref)
            logger.error(m)
            raise InheritableContentError(m)

    def is_resolved(self):
        unresolved = [ 1 for exmp in self.content.values()
                       if not exmp.is_resolved() ]

        if sum(unresolved) > 0:
            return False
        else:
            return True

    def resolve(self):
        for exmp in self.content.values():
            exmp.resolve(self.data)

class DataCache(RecursiveConfigurationBase):
    """
    Represents a group of related files that hold similar kinds of structured
    data. Often subclassed.
    """

    content_class = DataContentBase

    def __init__(self, files, conf):
        self._cache = {}
        self._conf = conf
        self.ingest(files)

    @property
    def cache(self):
        return self._cache

    @cache.setter
    def cache(self, value):
        logger.warning('cannot set cache record directly')

    def __contains__(self, key):
        return key in self.cache

    def _clear_cache(self, fn):
        self.cache[fn] = []

    def ingest(self, files):
        setup = [ self._clear_cache(fn)
                  for fn in files
                  if fn not in self.cache ]

        logger.debug('setup cache for {0} files'.format(len(setup)))

        for fn in files:
            self.add_file(fn)

    def add_file(self, fn):
        if fn not in self.cache or self.cache[fn] == []:
            data = ingest_yaml_list(fn)
            self.cache[fn] = self.content_class(data, self, self.conf)
        else:
            logger.info('populated file {0} exists in the cache'.format(fn))

    def fetch(self, fn, ref):
        if fn not in self.cache:
            logger.error('file "{0}" is not included.'.format(fn))
            raise InheritableContentError

        return self.cache[fn].fetch(ref)

class TitleData(ConfigurationBase):
    _option_registry = ['text']

    @property
    def level(self):
        if 'level' not in self.state:
            return 3
        else:
            return self.state['level']

    @level.setter
    def level(self, value):
        if isinstance(value, int):
            self.state['level'] = value
        else:
            raise TypeError
